import numpy as np
from scipy import optimize
from scipy.stats import norm
from statsmodels.distributions.empirical_distribution import ECDF
from statsmodels.stats.api import CompareMeans, DescrStatsW

from spearmint.config import (
    DEFAULT_ALPHA,
    DEFAULT_HYPOTHESIS,
    MIN_OBS_FOR_Z_TEST,
    logger,
)
from spearmint.table import SpearmintTable
from spearmint.typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Tuple,
    Union,
    Optional,
)
from spearmint.utils import format_value

SUPPORTED_POWER_STATISTICS = ("t", "z")


def _get_solve_power_function(statistic: str) -> Callable:
    if statistic not in SUPPORTED_POWER_STATISTICS:
        raise ValueError(f"Statistic '{statistic}' not supported for power calculation")

    if statistic == "t":
        from statsmodels.stats.power import tt_ind_solve_power as solve_power
    else:
        from statsmodels.stats.power import zt_ind_solve_power as solve_power
    return solve_power  # type: ignore


def bonferroni_correction(alpha_orig: float, p_values: Tuple[float]) -> float:
    """
    Calculate the correcteed alpha value for a frequentist hypothesis test using
    Bonferroni's method.

    en.wikipedia.org/wiki/Bonferroni_correction

    Parameters
    ----------
    alpha_orig : float
        alpha value before correction
    p_values: Tuple[float]
        The p-values associated with concurrent hypothesis tests being run

    Returns
    -------
    alpha_corrected: float
        new critical value (i.e. the corrected alpha)
    """
    n_tests = len(p_values)
    return alpha_orig / n_tests


def sidak_correction(alpha_orig: float, p_values: Tuple[float]) -> float:
    """
    Calculate the correcteed alpha value for a frequentist hypothesis test using
    Sidak's method.

    en.wikipedia.org/wiki/%C5%A0id%C3%A1k_correction

    Parameters
    ----------
    alpha_orig : float
        alpha value before correction
    p_values: Tuple[float]
        The p-values associated with concurrent hypothesis tests being run

    Returns
    -------
    alpha_corrected: float
        new critical value (i.e. the corrected alpha)
    """
    n_tests = len(p_values)
    return 1.0 - (1.0 - alpha_orig) ** (1.0 / n_tests)


def fdr_bh_correction(fdr: float, p_values: Tuple[float]) -> float:
    """
    Benjamini-Hochberg false-discovery rate adjustment procedure.

    pdfs.semanticscholar.org/af6e/9cd1652b40e219b45402313ec6f4b5b3d96b.pdf

    Parameters
    ----------
    fdr : float
        False Discovery Rate (q*), proportion of significant results that are
        actually false positives
    p_values: Iterable[float]
        The p-values associated with concurrent hypothesis tests being run

    Returns
    -------
    alpha_corrected: float
        The corrected alpha value
    """
    n_tests = len(p_values)

    def p_i(i):
        return i * fdr / n_tests

    p_sorted = np.sort(np.asarray(p_values))

    significant_idx = [i for i, val in enumerate(p_sorted, 1) if val <= p_i(i)]
    rank = np.max(significant_idx) if significant_idx else 1
    return p_i(rank)


MULTIPLE_COMPARISON_CORRECTION_METHODS = {
    "b": bonferroni_correction,
    "bonferroni": bonferroni_correction,
    "s": sidak_correction,
    "sidak": sidak_correction,
    "bh": fdr_bh_correction,
    "fdr_bh": fdr_bh_correction,
}


def _get_multiple_comparison_correction_function(
    method: Union[str, Callable]
) -> Callable:
    if isinstance(method, Callable):  # type: ignore
        if MULTIPLE_COMPARISON_CORRECTION_METHODS.values():
            return method  # type: ignore
    elif isinstance(method, str):
        if method in MULTIPLE_COMPARISON_CORRECTION_METHODS.keys():
            return MULTIPLE_COMPARISON_CORRECTION_METHODS[method]

    raise ValueError(f"Multiple correction method {method} not supported")


class MultipleComparisonCorrection:
    """
    Perform multiple comparison adjustment of alpha based on a sequence of
    p_values that result from two or more hypothesis tests inference procedures.

    p_values : Tuple[float]
        A list of p_values resulting from two or more hypothesis tests.
    method :  str
        One of the following correction methods:
            - 'bonferroni', 'b' : one-step Bonferroni correction
            - 'sidak', 's' : one-step Sidak correction
            - 'fdr_bh', 'bh; : Benjamini/Hochberg (non-negative)
    alpha : float in (0, 1)
        The desired probability of Type I error
    reject_null: list[bool]
        For each probablity, whether or not to reject the null hypothsis given
        the updated values for alpha.
    """

    __ATTRS__ = ["ntests", "method", "alpha_orig", "alpha_corrected"]

    def __init__(self, p_values, method="sidak", alpha=DEFAULT_ALPHA):
        self.mc_correction_method = _get_multiple_comparison_correction_function(method)
        self.ntests = len(p_values)
        self.alpha_orig = alpha
        self.alpha_corrected = self.mc_correction_method(alpha, p_values)
        self.accept_hypothesis = [p < self.alpha_corrected for p in p_values]

    @property
    def mc_correction_method_name(self) -> str:
        return self.mc_correction_method.__name__.split("_correction")[0]


def estimate_experiment_sample_sizes(
    delta: float,
    statistic: str = "z",
    alpha: float = 0.05,
    power: float = 0.8,
    *args,
    **kwargs,
) -> Tuple[int, ...]:
    """
    Calculate the sample size required for each treatement in order to observe a
    difference of `delta` between control and variation groups, given a
    particular `alpha` (Type I error rate) and `power` (1 - Type II error rate).

    Parameters
    ----------
    delta : float
        The absolute difference in means between control and variation groups
    statistic : string
        Either:
            - 'z' or 't' if interpreting effect size as scaled difference of means
            - 'rates_ratio' if interpereting effect size as the ratio of means
    alpha : float [0, 1)
            The assumed Type I error of the test
    power : float [0, 1)
        The desired statistical power of the test
    *args, **kwargs
        Model-specific arguments

    Returns
    -------
    sample_sizes : List[int]
        The estimated sample sizes for the control and variation treatments

    Example 1: Continuous Variables
    ----

    Estimate the sample size required to observe significant difference between
    two binomial distributions that differ by .01 in mean probability with
    Type I error = 0.05 (default) and Power = 0.8 (default)

    ```python
    >>> prob_control = .49
    >>> std_control = (prob_control * (1 - prob_control))**.5  # Binomial std
    >>> prob_variation = std_variation = .50
    >>> delta = prob_variation - prob_control
    >>> estimate_experiment_sample_sizes(
        delta=delta,
        statistic='z',
        std_control=std_control,
        std_variation=std_variation
    )
    (39236, 39236)
    ```python

    Example 2: Count Variables
    ---
    Replicate Example 1 from Gu et al, 2008

    ```python
    >>> R = 4  # ratio under alternative hypothesis
    >>> control_rate = .0005
    >>> variation_rate = R * control_rate
    >>> delta = variation_rate - control_rate
    >>> estimate_experiment_sample_sizes(
        delta,
        statistic='rates_ratio',
        control_rate=control_rate,
        alpha=.05,
        power=.9,
        control_exposure_time=2.,
        sample_size_ratio=.5
    )
    (8590, 4295)
    ```
    """
    if statistic in ("t", "z"):
        # std_control and/or std_variation are in *args, or **kwargs
        return cohens_d_sample_size(delta, alpha, power, statistic, *args, **kwargs)
    elif statistic == "rates_ratio":
        return ratio_sample_size(alpha, power, delta, *args, **kwargs)
    else:
        raise ValueError("Unknown statistic")


def cohens_d(delta: float, std_control: float, std_variation: Optional[float]) -> float:
    """Calculate the Cohen's d effect size comparing two samples. For details
    see https://en.wikiversity.org/wiki/Cohen%27s_d

    Parameters
    ----------
    delta : float
        The measured difference in means between the two samples
    std_control : float
        The standard deviation for the control sample
    std_variation : float, optional
        If provided, the variation for the treatment sample. If None provided,
        we assume the treatment has the same standard deviation as the control.

    Returns
    -------
    float
        Cohen'd d metric for effect size.
    """
    std_variation = std_variation if std_variation else std_control
    std_pooled = np.sqrt((std_control**2 + std_variation**2) / 2.0)
    return delta / std_pooled


def cohens_d_sample_size(
    delta: float,
    alpha: float,
    power: float,
    statistic: str,
    std_control: float,
    std_variation: Optional[float] = None,
    sample_size_ratio: float = 1.0,
) -> Tuple[int, int]:
    """
    Calculate sample size required to observe a significantly reliable difference
    between groups a and b. Assumes Cohen's d definition of effect size and an
    enrollment ratio of 1.0 between groups a and b by default.

    Parameters
    ----------
    std_control : float
        An estiamte of the expected sample standard deviation of control
        group
    nobs_control : int
        The number of control observations.
    std_variation : float
        An estimate of the expected sample standard deviation of variation
        group. If not provided, we assume homogenous variances for the
        two groups.

    Returns
    -------
    sample_sizes : list[int]
        The estimated sample sizes for the control and variation treatments

    Example
    -------
    Estimate of sample sizes required to observe a significant difference between
    two binomial distributions that have the same variance, but differ by .01
    in mean probability.

    ```python
    >>> mean_prob_control = .49
    >>> mean_prob_variation = std_variation = .5
    >>> std_control = (mean_prob_control * (1 - mean_prob_control))**.5  # Bernoulli stddev
    >>> std_variation = std_control
    >>> delta = mean_prob_variation - mean_prob_control
    >>> cohens_d_sample_size(
        delta=delta,
        alpha=.05,
        power=.8,
        statistic='z',
        std_control=std_control,
        std_variation=std_variation
    )
    (39228, 39228)
    ```

    References
    ----------
    Cohen, J. (1988). Statistical power analysis for the behavioral sciences
        (2nd ed.). Hillsdale, NJ: Lawrence Earlbaum Associates.
    """

    effect_size = cohens_d(delta, std_control, std_variation)

    solve_power_function = _get_solve_power_function(statistic)
    N1 = int(
        solve_power_function(
            effect_size, alpha=alpha, power=power, ratio=sample_size_ratio
        )
    )
    N2 = int(N1 * sample_size_ratio)
    return N1, N2


def ratio_sample_size(
    alpha: float,
    power: float,
    delta: float,
    control_rate: float,
    control_exposure_time: float = 1.0,
    null_ratio: float = 1.0,
    sample_size_ratio: float = 1.0,
    exposure_time_ratio: float = 1.0,
) -> Tuple[int, ...]:
    """
    Calculate sample size required to observe a significantly reliable ratio of
    rates between variation and control groups. Follows power calculation outlined
    in Gu et al, 2008.

    Parameters
    ----------
    control_rate : float
        The poisson rate of the control group
    control_exposure_time : float
        The number of time units of the control exposure. Default is 1.0
    null_ratio : float
        The ratio of variation to control rates under the null hypothesis.
        Default is 1.
    sample_size_ratio : float
        The ratio of sample sizes of the variation to the control groups. Default is
        1, thus assuming equal sample sizes.
    exposure_time_ratio : float
        The ratio of the variation exposure time to the control. Default is 1.0,
        thus assuming equal exposure times

    Returns
    -------
    sample_sizes : Tuple[int, int]
        Sample sizes for each group

    Example
    -------
    Replicate Example 1 from Gu et al, 2008

    ```python
    >>> R = 4  # ratio under alternative hypothesis
    >>> control_rate = .0005
    >>> variation_rate = R * control_rate
    >>> delta = variation_rate - control_rate
    >>> ratio_sample_size(
        alpha=.05,
        power=.9,
        delta=delta,
        control_rate=control_rate,
        control_exposure_time=2.,
        sample_size_ratio=.5
    )
    (8590, 4295)
    ```

    Returns (8590, 4295), which have been validated to be more accurate than
    the result reported in Gu et al, due to rounding precision. For details
    see "Example 2 – Validation using Gu (2008)" section of
    http://ncss-wpengine.netdna-ssl.com/wp-content/themes/ncss/pdf/Procedures/PASS/Tests_for_the_Ratio_of_Two_Poisson_Rates.pdf

    References
    ----------
    Gu, K., Ng, H.K.T., Tang, M.L., and Schucany, W. 2008. 'Testing the Ratio of
        Two Poisson Rates.' Biometrical Journal, 50, 2, 283-298.
    Huffman, M. 1984. 'An Improved Approximate Two-Sample Poisson Test.'
        Applied Statistics, 33, 2, 224-226.
    """

    # convert absolute difference to ratio
    alternative_ratio = float(control_rate + delta) / control_rate
    variation_exposure_time = exposure_time_ratio * control_exposure_time

    z_alpha = norm.ppf(1 - alpha)
    z_power = norm.ppf(power)

    def objective(x):
        ratio_proposed = (x[1] * variation_exposure_time) / (
            x[0] * control_exposure_time
        )
        loss = np.abs(null_ratio - (alternative_ratio / ratio_proposed))
        return loss

    def con1(x):
        """General sample size ratio constraint"""
        return (float(x[1]) / x[0]) - sample_size_ratio

    def con2(x):
        """Control sample size constraint, outlined in Gu et al, 2008, Equation 10"""

        N1, N2 = x
        d = (control_exposure_time * N1) / (variation_exposure_time * N2)
        A = 2 * (1.0 - np.sqrt(null_ratio / alternative_ratio))
        C = np.sqrt((null_ratio + d) / alternative_ratio)
        D = np.sqrt((alternative_ratio + d) / alternative_ratio)
        return x[0] - (((z_alpha * C + z_power * D) / A) ** 2.0 - (3.0 / 8)) / (
            control_exposure_time * control_rate
        )

    constraint1 = {"type": "eq", "fun": con1}
    constraint2 = {"type": "eq", "fun": con2}
    constraints = [constraint1, constraint2]

    results = optimize.minimize(
        objective,
        (10, 10),
        bounds=((1, None), (1, None)),
        constraints=constraints,
        method="SLSQP",
        tol=1e-10,
    )
    return tuple(int(np.ceil(n)) for n in results.x)


def highest_density_interval(
    samples: Iterable[float], credible_mass: float = 0.95
) -> np.ndarray:
    """
    Calculate the bounds of the highest density interval (HDI) with width
    `credible_mass` under the distribution of samples.

    Parameters
    ----------
    samples: Iterable[float]
        The samples to compute the interval over
    credible_mass: float in (0, 1)
        The credible mass under the empricial distribution

    Returns
    -------
    hdi: Tuple[float, float]
        The lower and upper bounds of the highest density interval

    Example
    -------
    >>> samples = np.random.randn(100000)  # standard normal samples
    >>> highest_density_interval(samples, credible_mass=0.95)  # should be approx +/- 1.96
    (-1.9600067573529922, 1.9562495685489785)
    """
    _samples = np.asarray(sorted(samples))
    n = len(_samples)

    interval_idx_inc = int(np.floor(credible_mass * n))
    n_intervals = n - interval_idx_inc
    interval_width = _samples[interval_idx_inc:] - _samples[:n_intervals]

    if len(interval_width) == 0:
        raise ValueError("Too few elements for interval calculation")

    min_idx = np.argmin(interval_width)
    hdi_min = _samples[min_idx]
    hdi_max = _samples[min_idx + interval_idx_inc]
    return np.array((hdi_min, hdi_max))


class EmpiricalCdf:
    """
    Class that calculates the empirical cumulative distribution function for a
    set of samples.
    """

    def __init__(self, samples: Iterable[float]):
        self.samples = samples
        self._cdf = ECDF(samples)

    @property
    def samples_cdf(self) -> ECDF:
        """
        Return the cdf evaluated at those samples used to estimate the cdf
        parameters.
        """
        if not hasattr(self, "_samples_cdf"):
            self._samples_cdf = self.evaluate(sorted(self.samples))
        return self._samples_cdf

    def __call__(self, values: Optional[Iterable[float]] = None) -> ECDF:
        """
        Callable interface for the EmpiricalCdf object

        Parameters
        ----------
        values: Iterable[float]
            Observations at which to evaluate the CDF.

        Returns
        -------
        ECDF: statsmodels.distributions.empirical_distribution.ECDF
            An empricial cumulative distribution funcion object. For details,
            see https://www.statsmodels.org/stable/generated/statsmodels.distributions.empirical_distribution.ECDF.html
        """
        return self.evaluate(values)

    def evaluate(self, values: Optional[Iterable[float]] = None) -> ECDF:
        """
        Evaluate the cdf for a sequence of values. If None provided, evaluate
        at all the `samples` used to estimate the CDF

        Parameters
        ----------
        values: Iterable[float]
            Observations at which to evaluate the CDF.

        Returns
        -------
        ECDF: statsmodels.distributions.empirical_distribution.ECDF
            An empricial cumulative distribution funcion object. For details,
            see https://www.statsmodels.org/stable/generated/statsmodels.distributions.empirical_distribution.ECDF.html
        """
        if values is None:
            return self.samples_cdf
        return self._cdf(values)


class Samples(DescrStatsW):
    """
    Class for holding samples and calculating various statistics on those
    samples. Any invalid observations (None, nan, or inf) are ignored.
    """

    def __init__(self, observations: Iterable[float], name: str = "Samples"):
        """
        Parameters
        ----------
        observations : Iterable[float]
            An array-like of observations
        name : str, optional
            A name used when displaying the sample statistics.

        Raises
        ------
        ValueError if all observations are None, nan, or Inf.
        """
        self.name = name
        self._raw_observations = observations
        self._summary_table = None
        valid_observations = self._valid_observations(observations)
        super(Samples, self).__init__(valid_observations)

    def _valid_observations(self, observations: Iterable[float]) -> np.ndarray:
        def valid(o):
            if o is None:
                return False
            if np.isnan(o):
                return False
            if np.isinf(0):
                return False
            return True

        observations = list(filter(valid, observations))
        if self.name:
            name_string = f"{self.name}"
        else:
            name_string = ""
        if not observations:
            raise ValueError(f"All {name_string} observations are nan or None")
        else:
            return np.array(observations)

    @property
    def summary(self) -> None:
        """Print a summary table for the valid observations in the sample"""
        if self._summary_table is None:
            self._summary_table = SamplesSummaryTable(self)  # type: ignore
        self._summary_table.print()  # type: ignore

    @property
    def max(self) -> float:
        return np.max(self.data)

    def permute(self) -> np.ndarray:
        """Shuffle the samples"""
        return np.random.choice(self.data, int(self.nobs))

    def sort(self) -> np.ndarray:
        """Return the observations in ascending order"""
        if not hasattr(self, "_sorted"):
            self._sorted = np.array(sorted(self.data))
        return self._sorted

    def percentiles(
        self, prct: Tuple[float, ...] = (2.5, 25.0, 50.0, 75.0, 97.5)
    ) -> np.ndarray:
        """Calculate the provided percentiles of the sample

        Parameters
        ----------
        prct : Iterable[float]
            Values in the range (0, 100) defining the percentiles to calcualte.

        Returns
        -------
        percentiles : np.ndarray
            The associated values in samples for the provided `prct` percentiles.
        """
        return np.percentile(self.data, prct)

    @property
    def cdf(self) -> EmpiricalCdf:
        """Return the empricial cumulative distribution function for the samples"""
        if not hasattr(self, "_cdf"):
            self._cdf = EmpiricalCdf(self.data)
        return self._cdf

    def prob_greater_than(
        self, values: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """
        Return the cumulative probability of `values` under the current samples'
        empirical CDF.
        """
        return 1.0 - self.cdf(np.asarray(values, dtype=float))

    def prob_less_than(
        self, values: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """
        Return the cumulative probability of `values` under the current samples'
        empirical CDF.
        """
        return self.cdf(np.asarray(values, dtype=float))

    def confidence_interval(self, confidence: float = 0.95) -> Tuple[float, float]:
        """
        Calculate the `confidence`-% confidence interval around the mean estimate,
        assuming a Gaussian approximation for the sample distribution.

        Parameters
        ----------
        confidence : float
            Confidence level for the interval.

        Returns
        -------
        ci : Tuple[float, float]
            The upper and lower bounds of the the `confidence`-% confidence interval
            around the mean estimate.
        """

        alpha = (1 - confidence) / 2.0  # symmetric about mean
        z = norm.ppf(1 - alpha)
        ci = z * self.std_err

        return self.mean - ci, self.mean + ci

    @property
    def std_err(self) -> Tuple[float, float]:
        return (self.var / self.nobs) ** 0.5

    def hdi(
        self, credible_mass: float = 0.95
    ) -> Union[Tuple[float, float], Tuple[None, None]]:
        """
        Calculate the bounds of the highest density interval (HDI) with width
        `credible_mass` under the distribution of samples.

        Parameters
        ----------
        credible_mass: float in (0, 1)
            The amount of probability under the returned HDI

        Returns
        -------
        hdi: Tuple[float, float]
            The boundary of the highest density interval for the sample
            distribution, (HDI_lower, HDI_upper)
        """
        try:
            hdi = highest_density_interval(self.data, credible_mass)
            # unpack
            return hdi[0], hdi[1]
        except Exception as e:
            logger.warning(e)
            return (None, None)


class SamplesSummaryTable(SpearmintTable):
    def __init__(self, samples: Samples):
        """Summary statistics comparing control and variation samples"""
        sample_name = samples.name if samples.name is not None else ""
        super().__init__(title="Samples Summary")

        self.add_column("", justify="right")
        self.add_column(sample_name)

        self.add_row(
            "Samples",
            format_value(samples.nobs),
        )
        self.add_row(
            "Mean",
            format_value(samples.mean),
        )
        self.add_row(
            "Standard Error",
            format_value(samples.std_err),
        )
        self.add_row(
            "Variance",
            format_value(samples.var),
        )


class SamplesComparisonTable(SpearmintTable):
    def __init__(
        self,
        control_samples: Samples,
        variation_samples: Union[Samples, List[Samples]],
        metric_name: str = "",
    ):
        """Summary statistics comparing control and variation samples"""
        super().__init__(title="Samples Comparison")

        if isinstance(variation_samples, Samples):
            variation_samples = [variation_samples]

        self.add_column(metric_name, justify="right")
        self.add_column(control_samples.name)

        for vs in variation_samples:
            self.add_column(vs.name)

        def get_variation_row_values(value, precision=4):
            "handle case when there are more than one variation conditions"
            return [
                format_value(getattr(vs, value), precision=precision)
                for vs in variation_samples
            ]

        self.add_row(
            "Samples",
            format_value(control_samples.nobs, precision=0),
            *get_variation_row_values("nobs", precision=0),
        )
        self.add_row(
            "Mean",
            format_value(control_samples.mean),
            *get_variation_row_values("mean", precision=4),
        )
        self.add_row(
            "Standard Error",
            format_value(control_samples.std_err),
            *get_variation_row_values("std_err", precision=4),
        )
        self.add_row(
            "Variance",
            format_value(control_samples.var),
            *get_variation_row_values("var", precision=4),
        )
        deltas = [None] + [
            format_value(vs.mean - control_samples.mean, precision=4)
            for vs in variation_samples
        ]
        self.add_row("Delta", *deltas)


class MeanComparison(CompareMeans):
    """
    Class for comparing the means of two continuous sample distributions,
    provides a number of helpful summary statistics about the comparison. Assumes
    the distributions can be approximated with a Gaussian distrubution.
    """

    def __init__(
        self,
        samples_a: Samples,
        samples_b: Samples,
        alpha: float = DEFAULT_ALPHA,
        test_statistic_name: str = "t",
        hypothesis: str = DEFAULT_HYPOTHESIS,
    ):
        """
        Parameters
        ----------
        samples_a : `Samples`
            Samples from group A (e.g. control group)
        samples_b : `Samples`
            Samples from group B (e.g. treatment group)
        alpha : float in (0, 1)
            The acceptable Type I error rate in the comparison
        test_statistic_name: str
            The name of the hypothesis test statistic used in the comparison.
                -   't': to use a t-test (small sample size, N <= 30)
                -   'z': to use a z-test (large samples size, N > 30)
        hypothesis : str
            Defines the assumed alternative hypothesis. Can be either:
                -   'larger': indicating we hypothesize that group B's mean is
                    larger than group A's
                -   'smaller': indicating we hypothesize that group B's mean is
                    smaller than group A's
                -   'unequal': indicating we hypothesize that group B's mean is
                    not equal to group A's (i.e. two-tailed test).
        """
        super().__init__(samples_a, samples_b)

        self.alpha = alpha
        self._test_statistic_name = test_statistic_name
        self.hypothesis = hypothesis
        self.warnings: List = []
        self._comparison_table = None

    @property  # property allows different test statistic name definitions
    def test_statistic_name(self) -> str:
        return self._test_statistic_name

    @property
    def pooled_variance(self) -> float:
        """The pooled variance of the the two groups"""
        return ((self.d2.nobs - 1) * self.d2.var + (self.d1.nobs - 1) * self.d1.var) / (
            self.d2.nobs + self.d1.nobs - 2
        )

    @property
    def delta(self) -> float:
        """The absolute difference between the group means"""
        return self.d1.mean - self.d2.mean

    @property
    def delta_relative(self) -> float:
        """The relative difference between the group means"""
        return (self.d1.mean - self.d2.mean) / np.abs(self.d2.mean)

    @property
    def effect_size(self) -> float:
        """The Cohen's d effect size of the comparison"""
        return self.delta / np.sqrt(self.pooled_variance)

    @property
    def test_direction(self) -> str:
        """
        The directionality of the hyopthesis test. Will either be 'larger',
        'smaller', or 'two-sided'
        """
        return self.hypothesis if self.hypothesis != "unequal" else "two-sided"

    @property
    def power(self) -> float:
        """
        The statistical power of the comparison (i.e. 1-beta of the comparison,
        where beta is the Type II error rate)
        """
        ratio = self.d1.nobs / self.d2.nobs

        solve_power_function = _get_solve_power_function(self.test_statistic_name)

        return solve_power_function(
            effect_size=self.effect_size,
            nobs1=self.d2.nobs,
            alpha=self.alpha,
            ratio=ratio,
            alternative=self.test_direction,
        )

    @property
    def comparison(self) -> None:
        """Displays a table summarizing the sample comparison comparison."""
        if self._comparison_table is None:
            self._comparison_table = SamplesComparisonTable(self.d1, self.d2)  # type: ignore
        self._comparison_table.print()  # type: ignore

    @property
    def t_test_stats(self) -> Dict[str, Any]:
        """Results for a t-test (small sample sizes)

        Returns
        -------
        test_stats : Dict[str, Any]
            The resulting test statistics, with the following structure:
            ```
            {
                "statistic_name": "t",
                "statistic_value": float,
                "p_value": float,
                "df": int,
                "hypothesis": str
            }
            ```
        """

        tstat, pval, degrees_freedom = self.ttest_ind(alternative=self.test_direction)
        return {
            "statistic_name": "t",
            "statistic_value": tstat,
            "degrees_freedom": degrees_freedom,
            "hypothesis": self.hypothesis,
            "p_value": pval,
            "alpha": self.alpha,
            "power": self.power,
        }

    @property
    def z_test_stats(self) -> Dict[str, Any]:
        """Results for a z-test (large sample sizes)

        Returns
        -------
        test_stats : Dict[str, Any]
            The resulting test statistics, with the following structure:
            ```
            {
                "statistic_name": "t",
                "statistic_value": float,
                "p_value": float,
                "df": int,
                "hypothesis": str
            }
            ```
        """
        tstat, pval = self.ztest_ind(alternative=self.test_direction)
        return {
            "statistic_name": "z",
            "statistic_value": tstat,
            "hypothesis": self.hypothesis,
            "p_value": pval,
            "alpha": self.alpha,
            "power": self.power,
        }


class ProportionComparison(MeanComparison):
    """
    Class for comparing the proportions of two sample distributions, provides a
    number of helpful summary statistics about the comparison. In order to use the
    z-distribution, we assume normality oF proportions and thus, by proxy, adequate
    sample sizes (e.g. N > 30).
    """

    def __init__(self, variance_assumption: str = "pooled", *args, **kwargs):
        """
        Parameters
        ----------
        samples_a : `Samples`
            Samples from group A (e.g. control group)
        samples_b : `Samples`
            Samples from group B (e.g. treatment group)
        alpha : float in (0, 1)
            The acceptable Type I error rate in the comparison
        hypothesis : str
            Defines the assumed alternative hypothesis. Can be either:
                -   'larger': indicating we hypothesize that group B's mean is
                    larger than group A's
                -   'smaller': indicating we hypothesize that group B's mean is
                    smaller than group A's
                -   'unequal': indicating we hypothesize that group B's mean is
                    not equal to group A's (i.e. two-tailed test).
        variance_assumption : str, optional
            The assumed variance assumption. If "pooled" (default), we calculate
            the pooled variance, otherwise we simply calculate the global
            variance across both samples
        """

        super().__init__(*args, **kwargs)
        self._test_statistic_name = "z"
        nobs = min(self.d1.nobs, self.d2.nobs)

        # to use Normal approx, must have "large" N
        if nobs < MIN_OBS_FOR_Z_TEST:
            warning = (
                "Normality assumption violated, > ",
                f"{MIN_OBS_FOR_Z_TEST} observations required.",
            )

            logger.warn(warning)
            self.warnings.append(warning)

        self.variance_assumption = variance_assumption

    @property
    def pooled_variance(self) -> float:
        if self.variance_assumption == "pooled":
            p1 = self.d1.mean
            p2 = self.d2.mean
            var1 = p1 * (1 - p1)
            var2 = p2 * (1 - p2)
            return ((self.d1.nobs - 1) * var1 + (self.d2.nobs - 1) * var2) / (
                self.d1.nobs + self.d2.nobs - 2
            )
        else:  # global variance
            p = np.mean(np.r_[self.d1.data, self.d2.data])
            return p * (1 - p)

    @property
    def z_test_stats(self) -> Dict[str, Any]:
        """Test statistics for proportions comparison on normal (z) test

        Returns
        -------
        test_stats: Dict[str, float]
            The resulting test statistics, with the following structure:
            ```
            {
                "statistic_name": str = "z",
                "statistic_value": float,
                "p_value": float,
                "hypothesis": str
            }
            ```
        """
        from statsmodels.stats.proportion import proportions_ztest

        prop_var = self.pooled_variance
        n_1 = self.d1.nobs
        s_1 = sum(self.d1.data)
        n_2 = self.d2.nobs
        s_2 = sum(self.d2.data)
        zstat, pval = proportions_ztest(
            [s_1, s_2], [n_1, n_2], alternative=self.test_direction, prop_var=prop_var
        )
        return {
            "statistic_name": "z",
            "statistic_value": zstat,
            "hypothesis": self.hypothesis,
            "p_value": pval,
            "alpha": self.alpha,
            "power": self.power,
        }


class RateComparison(MeanComparison):
    """
    Class for comparing the rates of two sample distributions, provides a number
    of helpful summary statistics about the comparison. Uses the exact conditional
    test based on binomial distribution, as described in Gu et al (2008)

    References
    ----------
    Gu, Ng, Tang, Schucany 2008: Testing the Ratio of Two Poisson Rates,
    Biometrical Journal 50 (2008) 2, 2008
    """

    def __init__(self, null_ratio: float = 1.0, *args, **kwargs):
        """
        Parameters
        ----------
        samples_a : `Samples`
                Samples from group A (e.g. control group)
        samples_b : `Samples`
            Samples from group B (e.g. treatment group)
        alpha : float in (0, 1)
        hypothesis : str
                Defines the assumed alternative hypothesis. Can be either:
                    -   'larger': indicating we hypothesize that group B's mean is
                        larger than group A's
                    -   'smaller': indicating we hypothesize that group B's mean is
                        smaller than group A's
                    -   'unequal': indicating we hypothesize that group B's mean is
                        not equal to group A's (i.e. two-tailed test).

        null_ratio : float, optional
            The ratio of the two samples underneath the null hypothesis. Default
            is 1.0, i.e. that the two samples are equivalent, and thus their
            ratio is unity.
        """
        super().__init__(*args, **kwargs)
        self._test_statistic_name = "W"
        self.null_ratio = null_ratio

    @property
    def rates_ratio(self) -> float:
        """
        Return the comparison ratio of the null rates ratio and the observed
        rates ratio.
        """
        actual_ratio = (self.d1.sum * self.d1.nobs) / (self.d2.sum * self.d2.nobs)
        return self.null_ratio / actual_ratio

    @property
    def delta(self) -> float:
        """
        Delta is the ratio of the variation to the control rates
        """
        return self.d1.mean / self.d2.mean

    @property
    def delta_relative(self) -> float:
        """Return the athe ratio of the variation to the control rates"""
        return self.delta

    @property
    def rates_test_stats(self) -> Dict[str, Any]:
        """
        Run the rates comparison hyptothesis test. Uses the W5 statistic defined
        in Gu et al., 2008

        Returns
        -------
        test_stats : Dict[str, Any]
            The resulting test statistics, with the following structure:
            ```
            {
                "statistic_name": "W",
                "statistic_value": float,
                "p_value": float,
                "hypothesis": str
            }
            ```
            The (W-statistic, p-value) of the test, where W-statistic is the W5
            statistic from Gu et al., 2008.
        """
        X1, X2 = self.d2.sum, self.d1.sum
        t1, t2 = self.d2.nobs, self.d1.nobs
        d = t1 / t2
        W = (
            2
            * (
                np.sqrt(X2 + (3.0 / 8))
                - np.sqrt((self.null_ratio / d) * (X1 + (3.0 / 8)))
            )
            / np.sqrt(1 + (self.null_ratio / d))
        )

        if self.hypothesis == "larger":
            pval = 1 - norm.cdf(W)
        elif self.hypothesis == "smaller":
            pval = norm.cdf(W)
        elif self.hypothesis == "unequal":
            pval = 1 - norm.cdf(abs(W))

        return {
            "statistic_name": "W",
            "statistic_value": W,
            "hypothesis": self.hypothesis,
            "p_value": pval,
            "alpha": self.alpha,
            "power": self.power,
        }

    @property
    def effect_size(self) -> float:
        """
        Effect size ranges from 0-1
        """
        return 1 - self.rates_ratio

    @property
    def power(self) -> float:
        """
        Return the statistical power of the current test. Follows the calculation
        from W statistic 5 in Gu et al., 2008
        """
        N2, t2 = self.d1.sum, self.d1.nobs
        N1, t1 = self.d2.sum, self.d2.nobs

        lambda_2, lambda_1 = np.abs(self.d1.mean), np.abs(self.d2.mean)
        alternative_ratio = np.abs(lambda_2 / lambda_1)
        z = norm.ppf(1 - self.alpha)

        d = float(t1 * N1) / (t2 * N2)

        A = np.abs(2.0 * (1.0 - np.sqrt(self.null_ratio / alternative_ratio)))
        B = np.sqrt(lambda_1 * t1 * N1 + (3.0 / 8))
        C = np.sqrt((self.null_ratio + d) / alternative_ratio)
        D = np.sqrt((alternative_ratio + d) / alternative_ratio)
        W = (A * B - z * C) / D

        return round(norm.cdf(W), 4)


class BootstrapStatisticComparison(MeanComparison):
    """
    Class for comparing a bootstrapped test statistic for two samples. Provides
    a number of helpful summary statistics about the comparison.

    References
    ----------
    Efron, B. (1981). "Nonparametric estimates of standard error: The jackknife,
    the bootstrap and other methods". Biometrika. 68 (3): 589-599
    """

    def __init__(
        self,
        n_bootstraps: int = 1000,
        statistic_function: Callable = np.mean,
        *args,
        **kwargs,
    ):
        """
        Parameters
        ----------
        samples_a : `Samples`
            Samples from group A (e.g. control group)
        samples_b : `Samples`
            Samples from group B (e.g. treatment group)
        alpha : float in (0, 1)
            The acceptable Type I error rate in the comparison
        hypothesis : str
            Defines the assumed alternative hypothesis. Can be either:
                -   'larger': indicating we hypothesize that group B's mean is
                    larger than group A's
                -   'smaller': indicating we hypothesize that group B's mean is
                    smaller than group A's
                -   'unequal': indicating we hypothesize that group B's mean is
                    not equal to group A's (i.e. two-tailed test).

        n_bootstraps : int
            The number of bootstrap samples to draw use for estimates.
        statistic_function : Callable, optional
            Function that returns a scalar test statistic when provided a sequence
            of samples. Defaults to the mean.
        *args, **kwargs:
            Valid arguments to `MeanComparison`
        """

        super().__init__(*args, **kwargs)

        self.n_bootstraps = n_bootstraps
        self.statistic_function = statistic_function

    @property
    def test_statistic_name(self) -> str:
        return f"bootstrap_{self.statistic_function.__name__}"

    @property
    def bootstrap_test_stats(self) -> Dict[str, Any]:
        """
        Run the sample comparison hyptothesis test. Uses the bootstrapped sample statistics

        Returns
        -------
        test_stats: Dict[str, Any]
            The resulting test statistics, with the following structure:
            ```
            {
                "statistic_name": "bootstrap_delta",
                "statistic_value": float,
                "statistic_function_name": str,
                "n_bootstraps": int,
                "p_value": float,
                "hypothesis": str
            }
            ```
        """
        all_samples = np.concatenate([self.d1.data, self.d2.data]).astype(float)

        d1_samples = np.random.choice(
            all_samples, (int(self.d1.nobs), self.n_bootstraps), replace=True
        )
        d1_statistics = np.apply_along_axis(
            self.statistic_function, axis=0, arr=d1_samples
        )

        d2_samples = np.random.choice(
            all_samples, (int(self.d2.nobs), self.n_bootstraps), replace=True
        )
        d2_statistics = np.apply_along_axis(
            self.statistic_function, axis=0, arr=d2_samples
        )

        control_bs_samples = np.random.choice(
            self.d2.data, (int(self.d2.nobs), self.n_bootstraps), replace=True
        )
        control_statistics = np.apply_along_axis(
            self.statistic_function, axis=0, arr=control_bs_samples
        )
        self.control_bootstrap = Samples(control_statistics, name="control")

        variation_bs_samples = np.random.choice(
            self.d1.data, (int(self.d1.nobs), self.n_bootstraps), replace=True
        )
        variation_statistics = np.apply_along_axis(
            self.statistic_function, axis=0, arr=variation_bs_samples
        )
        self.variation_bootstrap = Samples(variation_statistics, name="variation")

        # The null sampling distribution of test_statistic deltas
        self.null_dist = Samples(
            d2_statistics - d1_statistics, name=f"{self.test_statistic_name}-null"
        )

        if self.hypothesis == "larger":
            pval = 1 - self.null_dist.cdf(self.delta)  # type: ignore
        elif self.hypothesis == "smaller":
            pval = self.null_dist.cdf(self.delta)  # type: ignore
        elif self.hypothesis == "unequal":
            pval = 1 - self.null_dist.cdf(abs(self.delta))  # type: ignore

        return {
            "statistic_name": self.test_statistic_name,
            "statistic_value": self.delta,
            "statistic_function_name": self.statistic_function.__name__,
            "n_bootstraps": self.n_bootstraps,
            "hypothesis": self.hypothesis,
            "p_value": pval,
            "alpha": self.alpha,
            "power": self.power,
        }

    def delta_ci(self) -> Tuple:
        return self.confidence_interval(1 - self.alpha)

    def confidence_interval(self, confidence: float = 0.95) -> Tuple[float, float]:
        """
        Calculate the `confidence`-% confidence interval around the statistic delta.
        Uses bootstrapped approximation the statistic sampling distribution.

        Returns
        -------
        ci : Tuple[float, float]
            the `confidence`-% confidence interval around the statistic estimate.
        """
        alpha = 1 - confidence

        return self.deltas_dist.percentiles([100 * alpha, 100 * (1 - alpha)])  # type: ignore

    @property
    def deltas_dist(self) -> Samples:
        if not hasattr(self, "_deltas_dist"):
            d1_samples = np.random.choice(
                self.d1.data, (int(self.d1.nobs), self.n_bootstraps), replace=True
            )
            d1_statistics = np.apply_along_axis(
                self.statistic_function, axis=0, arr=d1_samples
            )

            d2_samples = np.random.choice(
                self.d2.data, (int(self.d2.nobs), self.n_bootstraps), replace=True
            )
            d2_statistics = np.apply_along_axis(
                self.statistic_function, axis=0, arr=d2_samples
            )

            self._deltas_dist = Samples(
                d1_statistics - d2_statistics, name=f"{self.test_statistic_name}-deltas"
            )

        return self._deltas_dist

    @property
    def delta(self) -> float:
        """
        Return the average difference in test statistic distributions
        """
        return self.deltas_dist.mean

    @property
    def delta_relative(self) -> float:
        """Return the average difference in test statistic distributions, as
        a percent change.
        """
        return self.delta / np.abs(self.statistic_function(self.d2.data))

    @property
    def power(self) -> float:
        """
        Return the statistical power of the current test.
        """

        # Need to run the bootstrap in order to obtain the power
        if not hasattr(self, "null_dist"):
            _ = self.bootstrap_test_stats

        critical_value = self.null_dist.percentiles(100.0 * (1 - self.alpha))  # type: ignore
        return self.deltas_dist.prob_greater_than(critical_value)  # type: ignore



import getpass
import logging
import os
from configparser import ConfigParser
from dataclasses import dataclass

import holoviews as hv

from spearmint.typing import Any, Iterable, Union
from spearmint.utils import coerce_value, mkdir

CONFIG_TEMPLATE = """
# Define default configuration for environment, including any common/standard
# treatments, enrollments, measures, and attributes.

# ----------------------- BEGIN TEMPLATE -----------------------

[core]
spearmint_home={SPEARMINT_HOME}

# Logging level
logging_level=INFO

# Defaults
[experiment]
default_measure_names=metric
enrollment=enrollment

[hypothesis_test]:
default_control_name=control
default_variation_name=variation
default_treatment_name=treatment
default_metric_name=metric
default_hypothesis=larger
default_inference_method=frequentist
default_alpha=.05
min_obs_for_z_test=30
default_attribute_names=attr_0,attr_1

[vis]
vis_backend=bokeh
figure_width_pixels=400
figure_height_pixels=400

[bayesian_inference]:
default_parameter_estimation_method=analytic
default_mcmc_sampler=nuts
n_posterior_samples=1000
"""

TEMPLATE_BEGIN_PATTERN = (
    "# ----------------------- BEGIN TEMPLATE -----------------------"
)


def expand_env_var(env_var: str) -> Any:
    """
    Expands (potentially nested) env vars by repeatedly applying
    `expandvars` and `expanduser` until interpolation stops having
    any effect.
    """
    if not env_var:
        return env_var
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


def render_config_template(template: str) -> str:
    """
    Generates a configuration from the provided template + variables defined in
    current scope

    Parameters
    ----------
    template : str
        A configuration template with interpolations `{{variables}}`

    Returns
    -------
    rendered_template : str
        The template with variables interpolated from environment, global, and
        local scope.
    """
    all_vars = {k: v for d in [globals(), locals()] for k, v in d.items()}
    return template.format(**all_vars)


# ConfigParser is "old-style" class which is a classobj, not a type.
# We thus use multiple inheritance with object to fix
class SpearmintConfigParser(ConfigParser, object):
    """
    Custom config parser, with some validations
    """

    def __init__(self, *args, **kwargs) -> None:
        super(SpearmintConfigParser, self).__init__(*args, **kwargs)
        self.is_validated = False

    def _validate(self) -> None:
        self.is_validated = True

    def read(self, filenames: Union[str, Iterable[str]]) -> None:  # type: ignore
        ConfigParser.read(self, filenames)
        self._validate()


# Home directory and configuration locations.
# We default to ~/.spearmint and ~/.spearmint/spearmint.cfg if not provided
if "SPEARMINT_HOME" not in os.environ:
    SPEARMINT_HOME = expand_env_var("~/.spearmint")
    os.environ["SPEARMINT_HOME"] = SPEARMINT_HOME
else:
    SPEARMINT_HOME = expand_env_var(os.environ["SPEARMINT_HOME"])

mkdir(SPEARMINT_HOME)

if "SPEARMINT_CONFIG" not in os.environ:
    if os.path.isfile(expand_env_var("~/spearmint.cfg")):
        SPEARMINT_CONFIG = expand_env_var("~/spearmint.cfg")
    else:
        SPEARMINT_CONFIG = SPEARMINT_HOME + "/spearmint.cfg"
else:
    SPEARMINT_CONFIG = expand_env_var(os.environ["SPEARMINT_CONFIG"])


if "SPEARMINT_USER" not in os.environ:
    SPEARMINT_USER = getpass.getuser()
    os.environ["SPEARMINT_USER"] = SPEARMINT_USER
else:
    SPEARMINT_USER = os.environ["SPEARMINT_USER"]

# If needed, write the config file
if not os.path.isfile(SPEARMINT_CONFIG):
    logging.info(f"Creating new spearment config file in: {SPEARMINT_CONFIG}")
    with open(SPEARMINT_CONFIG, "w") as f:
        cfg = render_config_template(CONFIG_TEMPLATE)
        f.write(cfg.split(TEMPLATE_BEGIN_PATTERN)[-1].strip())


CONFIG = SpearmintConfigParser()
CONFIG.read(SPEARMINT_CONFIG)


# Give this module get/set methods
def get(section: str, key: str, **kwargs) -> Any:
    """
    Retrieve properly-typed variables from config.

    Parameters
    ----------
    section : str
        The top-level section of the config to get
    key : str
        The specific config value to get

    Returns
    -------
    value : Any
        The propertly-typed configuration value

    Example
    -------
    from spearmint import config
    # print the currently-configure SPEARMINT_HOME directory
    print(config.get('core', 'spearmint_home'))
    """
    return coerce_value(CONFIG.get(section, key, **kwargs))


def set(section: str, option: str, value: Any) -> None:
    CONFIG.set(section, option, value)


# Configure Logging
logger = logging.getLogger(__name__)
logger.setLevel(get("core", "logging_level"))

# Configure defaults
DEFAULT_METRIC_NAME = get("hypothesis_test", "default_metric_name")
DEFAULT_TREATMENT_NAME = get("hypothesis_test", "default_treatment_name")
DEFAULT_CONTROL_NAME = get("hypothesis_test", "default_control_name")
DEFAULT_VARIATION_NAME = get("hypothesis_test", "default_variation_name")
DEFAULT_INFERENCE_METHOD = get("hypothesis_test", "default_inference_method")
DEFAULT_HYPOTHESIS = get("hypothesis_test", "default_hypothesis")
DEFAULT_ALPHA = get("hypothesis_test", "default_alpha")


MIN_OBS_FOR_Z_TEST = get("hypothesis_test", "min_obs_for_z_test")
# DEFAULT_BAYESIAN_INFERENCE_METHOD = get("pymc", "bayesian_inference_method")

# Configure Bayesian Inference
N_POSTERIOR_SAMPLES = get("bayesian_inference", "n_posterior_samples")
DEFAULT_PARAMETER_ESTIMATION_METHOD = get(
    "bayesian_inference", "default_parameter_estimation_method"
)


# Configure Visualization
# TODO: add these to default config
_DEFAULT_VIS_BACKEND = get("vis", "vis_backend")
FIGURE_WIDTH_PIXELS = get("vis", "figure_width_pixels")
FIGURE_HEIGHT_PIXELS = get("vis", "figure_height_pixels")


def _get_vis_backend():
    if _DEFAULT_VIS_BACKEND == "bokeh":
        try:
            import bokeh

            return bokeh.__name__
        except ImportError:
            logger.warning(
                "Bokeh not available, falling back to matplotlib visualizaiton backend"
            )
            return "matplotlib"

    return _DEFAULT_VIS_BACKEND


# Set the visualization backend
VIS_BACKEND = _get_vis_backend()
hv.extension(VIS_BACKEND)


def _get_figure_params():
    """Return backend-specific figure params"""

    def _get_figsize_kwargs():
        if VIS_BACKEND == "bokeh":
            return dict(width=FIGURE_WIDTH_PIXELS, height=FIGURE_HEIGHT_PIXELS)
        elif VIS_BACKEND == "matplotlib":
            import matplotlib
            from holoviews.plotting.mpl import MPLPlot

            MPLPlot.sublabel_format = ""

            pix_per_inch = matplotlib.rcParams["figure.dpi"]
            aspect_ratio = FIGURE_WIDTH_PIXELS / FIGURE_HEIGHT_PIXELS
            fig_inches = max([FIGURE_WIDTH_PIXELS, FIGURE_HEIGHT_PIXELS]) / pix_per_inch
            return {"aspect": aspect_ratio, "fig_inches": fig_inches}

    def _get_hover_kwargs():
        if VIS_BACKEND == "matplotlib":
            # No hover in matplotlib
            return {}
        return {"tools": ["hover"]}

    figure_kwargs = {}
    figure_kwargs.update(_get_figsize_kwargs())
    figure_kwargs.update(_get_hover_kwargs())

    return figure_kwargs


def _get_points_plot_params():
    if VIS_BACKEND == "matplotlib":
        return {"s": 50}
    return {"size": 5}


FIGURE_PARAMS = _get_figure_params()
POINTS_PLOT_PARAMS = _get_points_plot_params()


@dataclass
class COLORS:
    blue = "#4257B2"
    light_blue = "#A1C4FD"
    cyan = "#3CCFCF"
    green = "#388E34"
    light_green = "#28CC7D"
    dark_green = "#006060"
    yellow = "#FFCD1F"
    salmon = "#FF725B"
    red = "#FB3640"
    dark_red = "#AE2024"
    purple = "#8842C0"
    gray = "#687174"
    dark_gray = "#455357"
    light_gray = "#C0CACE"
    brown = "#665000"
    black = "#000000"



from abc import ABC, abstractmethod

import arviz as az
import numpy as np

from spearmint.stats import Samples
from spearmint.typing import Optional


class BayesianAnalyticModel(ABC):
    """Base class for Bayesian models that leverage analytic solutions based
    on conjugate priors to implement posterior updates.
    """

    def __init__(
        self,
        delta_param: str,
        control_name: Optional[str] = None,
        variation_name: Optional[str] = None,
        *args,
        **kwargs,
    ):
        self.delta_param = delta_param
        self.control_name = control_name
        self.variation_name = variation_name
        self._control_posterior = None
        self._variation_posterior = None
        self._prior = None

    @property
    def control_posterior(self):
        return self._control_posterior

    @property
    def variation_posterior(self):
        return self._variation_posterior

    @property
    def prior(self):
        return self._prior

    def sample(self, n_samples: int = 1000) -> az.InferenceData:
        """
        Draw samples from the posterior, and return the results as InferenceData

        Parameters
        ----------
        n_samples : int
            The number of posterior samples to draw

        Returns
        -------
        inference_data : InferenceData
            An Arviz InferenceData structure with a `.posterior` attribute.
        """
        # Sample the prior (for visualization)
        prior_samples = self.prior.rvs(size=n_samples)

        # Sample delta parameters from the posterior
        control_posterior_samples = self.control_posterior.rvs(size=n_samples)
        variation_posterior_samples = self.variation_posterior.rvs(size=n_samples)

        delta_posterior_samples = (
            variation_posterior_samples - control_posterior_samples
        )
        delta_relative_samples = (
            variation_posterior_samples - control_posterior_samples
        ) / np.abs(np.mean(control_posterior_samples))
        effect_size_posterior_samples = delta_posterior_samples / np.sqrt(
            np.var(control_posterior_samples) + np.var(variation_posterior_samples)
        )

        data_dict = {
            "prior": prior_samples,
            f"{self.delta_param}_control": control_posterior_samples,
            f"{self.delta_param}_variation": variation_posterior_samples,
            "delta": delta_posterior_samples,
            "delta_relative": delta_relative_samples,
            "effect_size": effect_size_posterior_samples,
        }

        return az.convert_to_inference_data(data_dict)

    @abstractmethod
    def calculate_posteriors(
        self, control_samples: Samples, variation_samples: Samples
    ) -> None:
        """
        Update the posterior distributions for the control and variation groups
        in light of observed samples

        Parameters
        ----------
        control_samples : Samples
            Observations for the control group
        variation_samples : Samples
            Observations for the variation group
        """
        pass



import pytest

from spearmint import config
from spearmint import Experiment, HypothesisTest
from spearmint.inference.bayesian.bayesian_inference import (
    UnsupportedParameterEstimationMethodException,
)
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def counts_data():
    return generate_fake_observations(
        distribution="poisson", n_treatments=3, n_observations=3 * 100
    )


@pytest.mark.pymc_test
def test_bayesian_counts_default(counts_data):
    exp = Experiment(data=counts_data)

    test = HypothesisTest(
        metric="metric", control="A", variation="C", inference_method="bayesian"
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "poisson"  # Default model for counts data
    assert (
        test_results.bayesian_parameter_estimation_method
        == config.DEFAULT_PARAMETER_ESTIMATION_METHOD
    )  # MCMC is default
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bayesian_poisson_ab_mcmc(counts_data):
    exp = Experiment(data=counts_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayesian_model_name="poisson",
        bayesian_parameter_estimation_method="mcmc",
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "poisson"  # Default model for counts data
    assert (
        test_results.bayesian_parameter_estimation_method == "mcmc"
    )  # MCMC is default
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bayesian_poisson_aa_mcmc(counts_data):
    exp = Experiment(data=counts_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        # bayesian_parameter_estimation_method="mcmc"  # MCMC is default
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "poisson"  # Default model for counts data
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


@pytest.mark.pymc_test
def test_poisson_aa_advi(counts_data):
    """
    ADVI parameter estimation not supported for discrete PDFs like the
    Poisson.
    """
    exp = Experiment(data=counts_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        variable_type="counts",
        inference_method="bayesian",
        bayesian_parameter_estimation_method="advi",
    )
    with pytest.raises(UnsupportedParameterEstimationMethodException):
        exp.run_test(test)


def test_poisson_ab_analytic(counts_data):
    """
    ADVI parameter estimation not supported for discrete PDFs like the
    Poisson.
    """
    exp = Experiment(data=counts_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        inference_method="bayesian",
        bayesian_parameter_estimation_method="analytic",
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert test_results.model_name == "poisson"  # Default model for counts data
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


def test_poisson_aa_analytic(counts_data):
    """
    ADVI parameter estimation not supported for discrete PDFs like the
    Poisson.
    """
    exp = Experiment(data=counts_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_parameter_estimation_method="analytic",
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert test_results.model_name == "poisson"  # Default model for counts data
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )



from collections import OrderedDict

import pytest

from spearmint.inference import InferenceProcedure, InferenceResults
from spearmint.stats import Samples
from spearmint.typing import DataFrame
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def test_samples():
    observations = generate_fake_observations(distribution="bernoulli")["metric"].values
    return Samples(observations=observations, name="test")


class ExtendedInferenceResults(InferenceResults):
    def __init__(self, alpha=0.05, *args, **kwargs):
        super().__init__(
            metric_name="test",
            delta=0.0,
            delta_relative=0.0,
            effect_size=0.0,
            hypothesis="larger",
            alpha=alpha,
            accept_hypothesis=False,
            inference_method="test_inference",
            variable_type="test_variable_type",
            warnings=["a warning"],
            visualization_function=lambda x: x,  # placeholder visualization function
            segmentation="test_segmentation",
            *args,
            **kwargs
        )

    @property
    def _specific_properties(self) -> OrderedDict:
        return OrderedDict([("specific_property", "blah")])

    def _render_stats_table(self):
        class DummyStatsTable:
            """placeholder stats table with print method"""

            def print(self):
                pass

        return DummyStatsTable()


class ExtendedInferenceProcedure(InferenceProcedure):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inference_run = False

    def _run_inference(
        self, control_samples, variation_samples, **inference_kwargs
    ) -> None:
        self.inference_run = True
        self.control_samples = control_samples
        self.variation_samples = variation_samples

    def _make_results(self) -> InferenceResults:
        return ExtendedInferenceResults(
            control=self.control_samples,
            variation=self.variation_samples,
            alpha=self.alpha,
        )


def validate_results(results, alpha: float = 0.05):
    assert results.dict["control_name"] == results.dict["variation_name"]
    assert results.dict["control_nobs"] == results.dict["variation_nobs"]
    assert results.dict["control_mean"] == results.dict["variation_mean"]
    assert results.dict["control_ci"] == results.dict["variation_ci"]
    assert results.dict["control_var"] == results.dict["variation_var"]
    assert results.dict["metric"] == "test"
    assert results.dict["delta"] == 0
    assert results.dict["delta_relative"] == 0
    assert results.dict["effect_size"] == 0
    assert results.dict["alpha"] == alpha
    assert results.dict["hypothesis"] == "larger"
    assert results.dict["accept_hypothesis"] is False
    assert results.dict["inference_method"] == "test_inference"
    assert results.dict["variable_type"] == "test_variable_type"
    assert results.dict["warnings"] == "a warning"
    assert results.dict["specific_property"] == "blah"


def test_inference_results(test_samples):
    results = ExtendedInferenceResults(control=test_samples, variation=test_samples)
    validate_results(results)
    results.summary

    # Check datarame export support
    results_df = results.to_dataframe()
    assert isinstance(results_df, DataFrame)

    # We added one _specific property
    assert len(results_df.columns) == len(results._base_properties) + 1


def test_inference_procedure(test_samples):
    inference_procedure = ExtendedInferenceProcedure(inference_method="test", alpha=0.1)
    inference_results = inference_procedure.run(
        control_samples=test_samples, variation_samples=test_samples
    )
    assert inference_procedure.inference_run
    assert inference_procedure.alpha == 0.1
    assert inference_results == inference_procedure.results
    validate_results(inference_procedure.results, inference_procedure.alpha)

    inference_procedure.results.summary



import pymc as pm
from scipy import stats

from spearmint.stats import Samples
from spearmint.typing import Dict, Tuple, Optional

from .analytic_base import BayesianAnalyticModel


class BinomialAnalyticModel(BayesianAnalyticModel):
    """
    Implement analytic posterior updates for Beta-Binomial model. A Beta prior
    and Binomial likelihood result in a Beta posterior over proportionality
    parameters, which can be calculated efficiently from the prior and descriptive
    statistics of the observations.

    Notes
    -----
    We use the same model for Bernoulli and Binomial Likelihoods, as their
    posterirs updates are calculated in the same way.

    References
    ----------
    - https://www.cs.ubc.ca/~murphyk/Teaching/CS340-Fall06/lectures/bernoulli-slides.pdf
    """

    def __init__(
        self, prior_alpha: float = 1, prior_beta: float = 1.0, *args, **kwargs
    ):
        """
        Parameters
        ----------
        prior_alpha : float, optional
            The shape parameter for the Beta prior distribution
        prior_beta : float, optional
            The shape parametrer for the Beta prior distribution
        """
        super().__init__(delta_param="p", *args, **kwargs)  # type: ignore # (mypy bug, see #6799)
        self.prior_alpha = prior_alpha
        self.prior_beta = prior_beta

    def calculate_posteriors(
        self, control_samples: Samples, variation_samples: Samples
    ) -> None:
        """
        Update the posterior distributions for the control and variation in
        light of Samples

        Parameters
        ----------
        control_samples : Samples
            Observations for the control group
        variation_samples : Samples
            Observations for the variation group
        """

        # Update rules
        def _posterior_alpha(prior_alpha, samples):
            n_success = samples.sum
            return prior_alpha + n_success

        def _posterior_beta(prior_beta, samples):
            n = samples.nobs
            n_fail = n - samples.sum

            return prior_beta + n_fail

        control_posterior_alpha = _posterior_alpha(self.prior_alpha, control_samples)
        control_posterior_beta = _posterior_beta(self.prior_beta, control_samples)

        variation_posterior_alpha = _posterior_alpha(
            self.prior_alpha, variation_samples
        )
        variation_posterior_beta = _posterior_beta(self.prior_beta, variation_samples)

        self._control_posterior = stats.beta(
            control_posterior_alpha, control_posterior_beta
        )

        self._variation_posterior = stats.beta(
            variation_posterior_alpha, variation_posterior_beta
        )

        # Add prior for visualization
        self._prior = stats.beta(self.prior_alpha, self.prior_beta)


def build_binomial_analytic_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_alpha: float = 1.0,
    prior_beta: float = 1.0,
) -> Tuple[BinomialAnalyticModel, Dict[str, float]]:
    model = BinomialAnalyticModel(prior_alpha=prior_alpha, prior_beta=prior_beta)
    model.calculate_posteriors(control_samples, variation_samples)

    hyperparams = {"prior_alpha": prior_alpha, "prior_beta": prior_beta}
    return model, hyperparams


def build_bernoulli_analytic_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_alpha: float = 1.0,
    prior_beta: float = 1.0,
) -> Tuple[BinomialAnalyticModel, Dict[str, float]]:
    """Note: we use the same model for Binomial/Bernoulli, as they are equivalent"""
    return build_binomial_analytic_model(
        control_samples=control_samples,
        variation_samples=variation_samples,
        prior_alpha=prior_alpha,
        prior_beta=prior_beta,
    )


def build_bernoulli_pymc_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_alpha: float = 1.0,
    prior_beta: float = 1.0,
) -> pm.Model:
    """
    Compiles a Beta-Bernoulli Bayesian PyMC model for modeling binary data. The
    model consists of a Beta prior over proportionality (conversion rate), and
    a Bernoulli likelihood. For the Beta prior, we derive parameters separately
    for the control and variation from their observations (i.e. no pooling).

    Parameters
    ----------
    control_observations: Samples, dtype=int
        The control group observations
    variation_observations: Samples, dtype=int
        The variation group observations

    Returns
    -------
    model : pm.Model
        The compiled PyMC model built with the provided data
    hyperparams : Dict
        The prior parameters derived from the Samples

    Notes
    -----
    The Beta-Bernoulli model is equivalent to the Beta-Binomial model, and should
    provide very similar estiamtes of proportionality and dleta, but the
    Beta-Binomial model tends to be more computationally efficient.

    References
    ----------
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Beta.html
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Bernoulli.html
    """

    hyperparams = {"prior_alpha": prior_alpha, "prior_beta": prior_beta}

    with pm.Model() as model:
        # Priors
        pm.Beta("prior", alpha=prior_alpha, beta=prior_beta)  # for displaying prior
        p_control = pm.Beta("p_control", alpha=prior_alpha, beta=prior_beta)
        p_variation = pm.Beta("p_variation", alpha=prior_alpha, beta=prior_beta)

        # Likelihoods
        pm.Bernoulli("control", p=p_control, observed=control_samples.data)
        pm.Bernoulli("variation", p=p_variation, observed=variation_samples.data)

        # Inference parameters
        delta = pm.Deterministic("delta", p_variation - p_control)
        pm.Deterministic("delta_relative", (p_variation / p_control) - 1.0)
        pm.Deterministic(
            "effect_size",
            delta
            / pm.math.sqrt(
                (p_control * (1 - p_control) + p_variation * (1 - p_variation)) / 2
            ),
        )

    return model, hyperparams


def build_binomial_pymc_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_alpha: float = 1.0,
    prior_beta: float = 1.0,
    possible_outcomes: Optional[float] = None,
) -> pm.Model:
    """
    Compiles a Beta-Binomial Bayesian PyMC model for modeling binary data. The
    model consists of a Beta prior over proportionality (conversion rate), and
    a Binomial likelihood. For the Beta prior, we derive parameters separately
    for the control and variation from their observations (i.e. no pooling).

    Parameters
    ----------
    control_observations: Samples, dtype=int
        The control group observations
    variation_observations: Samples, dtype=int
        The variation group observations
    prior_alpha: float
        The location parameter for the Beta prior distribution over proportions
    prior_beta: float
        The shape parameter for the Beta prior distribution over proportions
    possible_outcomes: int
        The maximum number of possible outcomes for each trial. If None provided,
        we estimate it from the data.

    Returns
    -------
    model : pm.Model
        The compiled PyMC model built with the provided data
    hyperparams : Dict
        The prior parameters derived from the Samples

    Notes
    -----
    The Beta-Binomial model is equivalent to the Beta-Bernoulli model, and should
    provide very similar estiamtes of proportionality and dleta, but the
    Beta-Binomial model tends to be more computationally efficient.

    References
    ----------
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Beta.html
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Binomial.html
    """

    possible_outcomes = (
        max((control_samples.max, variation_samples.max))
        if possible_outcomes is None
        else possible_outcomes
    )
    hyperparams = {
        "prior_alpha": prior_alpha,
        "prior_beta": prior_beta,
        "possible_outcomes": possible_outcomes,
    }

    with pm.Model() as model:
        # Priors
        pm.Beta("prior", alpha=prior_alpha, beta=prior_beta)  # for displaying prior
        p_control = pm.Beta("p_control", alpha=prior_alpha, beta=prior_beta)
        p_variation = pm.Beta("p_variation", alpha=prior_alpha, beta=prior_beta)

        # Likelihoods
        pm.Binomial(
            "control",
            p=p_control,
            n=possible_outcomes,
            observed=control_samples.data,
        )
        pm.Binomial(
            "variation",
            p=p_variation,
            n=possible_outcomes,
            observed=variation_samples.data,
        )

        # Inference parameters
        delta = pm.Deterministic("delta", p_variation - p_control)
        pm.Deterministic("delta_relative", (p_variation / p_control) - 1.0)
        pm.Deterministic(
            "effect_size",
            delta
            / pm.math.sqrt(
                (p_control * (1 - p_control) + p_variation * (1 - p_variation)) / 2
            ),
        )

    return model, hyperparams



import logging
import os
import tempfile

import pytest

TEST_ENV_VARS = {"SPEARMINT_USER": "test", "SPEARMINT_HOME": tempfile.mkdtemp()}


@pytest.fixture(scope="session", autouse=True)
def test_config():
    original_env = dict(os.environ)

    # Set update test env
    os.environ.update(TEST_ENV_VARS)

    import importlib

    from spearmint import config

    # Reload the config with updated env vars
    importlib.reload(config)

    yield config

    # Reset the env to original state
    os.environ.clear()
    os.environ.update(original_env)


def test_env_vars(test_config):
    assert TEST_ENV_VARS["SPEARMINT_HOME"] == test_config.SPEARMINT_HOME
    assert TEST_ENV_VARS["SPEARMINT_USER"] == test_config.SPEARMINT_USER


def test_config_file(test_config):
    assert os.path.isfile(test_config.SPEARMINT_CONFIG)


def test_logger_level(test_config):
    assert test_config.logger.level == getattr(
        logging, (test_config.get("core", "logging_level"))
    )


def test_expand_env_var(test_config):
    assert "blammo" == test_config.expand_env_var("blammo")


def test_render_config_template(test_config):
    template = "{SPEARMINT_USER}"
    assert test_config.render_config_template(template) == test_config.SPEARMINT_USER


def test_get_set(test_config):
    test_config.set("core", "test_value", "test")
    assert test_config.get("core", "test_value") == "test"
    assert test_config.CONFIG.get("core", "test_value") == "test"



import pymc as pm
from scipy import stats

from spearmint.inference.bayesian.bayesian_inference import (
    UnsupportedParameterEstimationMethodException,
)
from spearmint.stats import Samples
from spearmint.typing import Dict, Tuple

from .analytic_base import BayesianAnalyticModel


class GaussianAnalyticModel(BayesianAnalyticModel):
    """
    Implement analytic posterior updates for hierarchical Gaussian model. A
    Gaussian prior and likelihood result in a Gaussian posterior over mean
    parameters, which can be calculated efficiently from the prior and
    descriptive statistics of the observations.

    References
    ----------
    - https://www.cs.ubc.ca/~murphyk/Papers/bayesGauss.pdf
    """

    def __init__(self, prior_mean: float = 0, prior_var: float = 1.0, *args, **kwargs):
        """
        Parameters
        ----------
        prior_mean : float, optional
            The mean of the prior distribution means, by default 0
        prior_var : float, optional
            The variance of the prior distribution on means, by default 1.0
        """
        super().__init__(delta_param="mu", *args, **kwargs)  # type: ignore # (mypy bug, see #6799)
        self.prior_mean = prior_mean
        self.prior_var = prior_var

    def calculate_posteriors(
        self, control_samples: Samples, variation_samples: Samples
    ) -> None:
        """
        Update the posterior distributions for the control and variation in
        light of Samples

        Parameters
        ----------
        control_samples : Samples
            Observations for the control group
        variation_samples : Samples
            Observations for the variation group
        """

        # Update rules
        def _posterior_variance(prior_var, samples):
            n = samples.nobs
            sample_var = samples.var

            return 1 / ((n / sample_var) + (1 / prior_var))

        def _posterior_mean(prior_mean, prior_var, posterior_var, samples):
            n = samples.nobs
            sample_mean = samples.mean
            sample_var = samples.var
            return posterior_var * (
                prior_mean / prior_var + n * sample_mean / sample_var
            )

        control_posterior_var = _posterior_variance(self.prior_var, control_samples)
        control_posterior_mean = _posterior_mean(
            self.prior_mean, self.prior_var, control_posterior_var, control_samples
        )

        variation_posterior_var = _posterior_variance(self.prior_var, variation_samples)
        variation_posterior_mean = _posterior_mean(
            self.prior_mean,
            self.prior_var,
            variation_posterior_var,
            variation_samples,
        )
        self._control_posterior = stats.norm(
            control_posterior_mean, control_posterior_var**0.5
        )

        self._variation_posterior = stats.norm(
            variation_posterior_mean, variation_posterior_var**0.5
        )
        # Add prior samples for visualization
        self._prior = stats.norm(self.prior_mean, self.prior_var)


def build_gaussian_analytic_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_mean: float = 0.0,
    prior_var: float = 5.0,
) -> Tuple[GaussianAnalyticModel, Dict[str, float]]:
    model = GaussianAnalyticModel(prior_mean=prior_mean, prior_var=prior_var)
    model.calculate_posteriors(control_samples, variation_samples)

    hyperparams = {"prior_mean_mu": prior_mean, "prior_var_mu": prior_var}
    return model, hyperparams


def build_gaussian_pymc_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_mean_mu: float = 0.0,
    prior_var_mu: float = 5.0,
    prior_mean_sigma: float = 1.0,
    prior_var_sigma: float = 5.0,
) -> Tuple[pm.Model, Dict[str, float]]:
    """
    Compiles a Hierarchical Gaussian Bayesian PyMC model for modeling binary data.
    The model consists of a Gaussian data likelihood with a Gaussian prior over
    the means, and Half-Normal prior over variance. For the priors, we derive
    separate location parameters for the control and variation from their
    associated observations (i.e. no pooling).

    Parameters
    ----------
    control_observations: Sample, dtype=float
        The control group observations
    variation_observations: Samples, dtype=float
        The variation group observations

    Returns
    -------
    model : pm.Model
        The compiled PyMC model built with the provided data
    hyperparams : Dict
        The prior parameters derived from the Samples

    References
    ----------
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Guassian.html
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.TruncatedNormal.html
    """

    with pm.Model() as model:
        hyperparams = {
            "prior_mean_mu": prior_mean_mu,
            "prior_var_mu": prior_var_mu,
            "prior_mean_sigma": prior_mean_sigma,
            "prior_var_sigma": prior_var_sigma,
        }

        # Priors
        sigma_control = pm.TruncatedNormal(
            "sigma_control", lower=1e-4, mu=prior_mean_sigma, sigma=prior_var_sigma
        )
        sigma_variation = pm.TruncatedNormal(
            "sigma_variation", lower=1e-4, mu=prior_mean_sigma, sigma=prior_var_sigma
        )
        pm.Normal(
            "prior", mu=prior_mean_mu, sigma=prior_var_mu
        )  # for visualizing prior
        mu_control = pm.Normal("mu_control", mu=prior_mean_mu, sigma=prior_var_mu)
        mu_variation = pm.Normal("mu_variation", mu=prior_mean_mu, sigma=prior_var_mu)

        # Likelihoods
        pm.Normal(
            "control", mu=mu_control, sigma=sigma_control, observed=control_samples.data
        )
        pm.Normal(
            "variation",
            mu=mu_variation,
            sigma=sigma_variation,
            observed=variation_samples.data,
        )

        # Inference parameters
        delta = pm.Deterministic("delta", mu_variation - mu_control)
        pm.Deterministic(
            "delta_relative",
            (mu_variation - mu_control) / pm.math.abs(control_samples.mean),
        )
        pm.Deterministic(
            "effect_size",
            delta / pm.math.sqrt((sigma_control**2.0 + sigma_variation**2.0) / 2.0),
        )

    return model, hyperparams


def build_student_t_analytic_model(
    control_samples: Samples,
    variation_samples: Samples,
) -> None:
    raise UnsupportedParameterEstimationMethodException(
        "Analytic parameter estimation not supported for `student_t` model"
    )


def build_student_t_pymc_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_mean_mu: float = 0.0,
    prior_var_mu: float = 5.0,
    prior_mean_sigma: float = 1.0,
    prior_var_sigma: float = 5.0,
    prior_nu_precision: float = 0.5,
) -> Tuple[pm.Model, Dict[str, float]]:
    """
    Compiles a Hierarchical Student-t Bayesian PyMC model for modeling binary data.
    Using the Student-t model allows for robust inference in the presence of
    outliers. The model consists of a Student-t data likelihood, a Gaussian
    prior over the mean, a Half-Normal prior over variance, and an Exponential
    distribution over the degress of freedom. For the priors, we derive separate
    location parameters for the control and variation from their associated
    observations (i.e. no pooling).

    Parameters
    ----------
    control_observations: Sample, dtype=float
        The control group observations
    variation_observations: Samples, dtype=float
        The variation group observations

    Returns
    -------
    model : pm.Model
        The compiled PyMC model built with the provided data
    hyperparams : Dict
        The prior parameters derived from the Samples

    References
    ----------
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.StudentT.html
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Guassian.html
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.TruncatedNormal.html
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Exponential.html
    """

    with pm.Model() as model:
        # Priors
        sigma_control = pm.TruncatedNormal(
            "sigma_control", lower=1e-4, mu=prior_mean_sigma, sigma=prior_var_sigma
        )
        sigma_variation = pm.TruncatedNormal(
            "sigma_variation", lower=1e-4, mu=prior_mean_sigma, sigma=prior_var_sigma
        )
        pm.Normal(
            "prior", mu=prior_mean_mu, sigma=prior_var_mu
        )  # for visualizing prior
        mu_control = pm.Normal("mu_control", mu=prior_mean_mu, sigma=prior_var_mu)
        mu_variation = pm.Normal("mu_variation", mu=prior_mean_mu, sigma=prior_var_mu)

        nu_control = pm.Exponential("nu_control", lam=1 / prior_nu_precision)
        nu_variation = pm.Exponential("nu_variation", lam=1 / prior_nu_precision)

        # Likelihoods
        pm.StudentT(
            "control",
            nu=nu_control,
            mu=mu_control,
            sigma=sigma_control,
            observed=control_samples.data,
        )
        pm.StudentT(
            "variation",
            nu=nu_variation,
            mu=mu_variation,
            sigma=sigma_variation,
            observed=variation_samples.data,
        )

        # Inference parameters
        delta = pm.Deterministic("delta", mu_variation - mu_control)
        pm.Deterministic(
            "delta_relative",
            (mu_variation - mu_control) / pm.math.abs(control_samples.mean),
        )
        pm.Deterministic(
            "effect_size",
            delta / pm.math.sqrt((sigma_control**2.0 + sigma_variation**2.0) / 2.0),
        )

    hyperparams = {
        "prior_mean_mu": prior_mean_mu,
        "prior_var_mu": prior_var_mu,
        "prior_mean_sigma": prior_mean_sigma,
        "prior_var_sigma": prior_var_sigma,
        "prior_nu_precision": prior_nu_precision,
    }
    return model, hyperparams



import numpy as np

from spearmint.stats import BootstrapStatisticComparison, Samples
from spearmint.typing import Callable, FilePath, Optional

from .frequentist_inference import (
    FrequentistInferenceProcedure,
    FrequentistInferenceResults,
)


def visualize_bootstrap_delta_results(
    results: FrequentistInferenceResults, outfile: Optional[FilePath] = None
):  # pragma: no cover
    # Lazy import
    import holoviews as hv

    from spearmint import vis

    test_statistic_label = results.test_statistic_name.replace("_", " ")
    test_statistic_title = test_statistic_label.title()
    control_samples = results.aux["control_bootstrap_samples"]
    variation_samples = results.aux["variation_bootstrap_samples"]
    delta_samples = results.aux["delta_bootstrap_samples"]

    ci_percent = round(100 * (1 - results.alpha))
    ci_bounds = np.round(100 * np.array((results.alpha / 2, 1 - results.alpha / 2)))

    control_dist = vis.plot_kde(
        samples=control_samples.data,
        label=results.control.name,
        color=vis.CONTROL_COLOR,
    )

    variation_dist = vis.plot_kde(
        samples=variation_samples.data,
        label=results.variation.name,
        color=vis.VARIATION_COLOR,
    )

    # Confidence intervals
    control_ci = vis.plot_interval(
        *control_samples.percentiles(ci_bounds),
        middle=control_samples.mean,
        label=results.control.name,
        color=vis.CONTROL_COLOR,
        show_interval_text=True,
    )  # type: ignore # (mypy bug, see #6799)

    variation_ci = vis.plot_interval(
        *variation_samples.percentiles(ci_bounds),
        middle=variation_samples.mean,
        label=results.variation.name,
        color=vis.VARIATION_COLOR,
        show_interval_text=True,
    )  # type: ignore # (mypy bug, see #6799)

    distribution_plot = control_dist * variation_dist * control_ci * variation_ci
    distribution_plot = distribution_plot.relabel(
        f"{test_statistic_title} Comparison"
    ).opts(legend_position="top_right", xlabel=test_statistic_label, ylabel="pdf")

    delta_dist = vis.plot_kde(
        samples=delta_samples.data,
        label="Delta Distribution",
        color=vis.DELTA_COLOR,
    )

    max_pdf_height = delta_dist.data["pdf"].max()

    mean_delta = results.aux["delta_bootstrap_samples"].mean

    delta_ci = vis.plot_interval(
        *delta_samples.percentiles(ci_bounds),
        mean_delta,
        color=vis.NEUTRAL_COLOR,
        label=f"{ci_percent}% Confidence Interval",
        show_interval_text=True,
        vertical_offset=-(max_pdf_height * 0.01),
    )  # type: ignore # (mypy bug, see #6799)

    zero_delta_vline = hv.Spikes(
        ([0.0], [max_pdf_height]), vdims="pdf", label="Null Delta"
    ).opts(color=vis.COLORS.red)

    delta_plot = delta_dist * delta_ci * zero_delta_vline
    delta_plot_title = f"{test_statistic_title} Delta"
    delta_plot = (
        delta_plot.relabel(delta_plot_title)
        .opts(xlabel="delta", ylabel="pdf")
        .opts(legend_position="top_right")
    )

    visualization = distribution_plot + delta_plot
    visualization.opts(shared_axes=False)
    # visualization.opts(shared_axes=False).cols(1)

    if outfile is not None:
        vis.save_visualization(visualization, outfile)

    return visualization


class BootstrapDelta(FrequentistInferenceProcedure):
    """
    Runs frequentist inference procedure to test for the difference in a bootstrapped
    test statistic estimate
    """

    def __init__(self, statistic_function: Callable = np.mean, *args, **kwargs):
        """
        Parameters
        ----------
        statistic_function : Callable
            Function that returns a scalar test statistic when provided a sequence
            of samples.

        """
        super(BootstrapDelta, self).__init__(*args, **kwargs)
        self.statistic_function = statistic_function

    @property
    def test_stats(self):
        return self.comparison.bootstrap_test_stats

    @property
    def delta_ci(self):
        """
        Calculate confidence interval around deltas with percentiles and values.
        """
        return self.comparison.confidence_interval(1 - self.alpha)

    # @abstractmethod
    def _run_inference(
        self, control_samples: Samples, variation_samples: Samples, **inference_kwargs
    ) -> None:
        self.comparison = BootstrapStatisticComparison(
            samples_a=variation_samples,
            samples_b=control_samples,
            alpha=self.alpha,
            hypothesis=self.hypothesis,
            statistic_function=self.statistic_function,
        )

    def accept_hypothesis(self, statistic_value: float) -> bool:
        """
        Overloads `FrequentistInferenceProcedure.accept_hypothesis()` method
        to use the boostrapped estimate of the sampling distribition to test the
        Null hypothesis
        """
        if self.hypothesis == "larger":
            return self.alpha > self.comparison.null_dist.prob_greater_than(
                statistic_value
            )
        elif self.hypothesis == "smaller":
            return self.alpha > 1 - self.comparison.null_dist.prob_greater_than(
                statistic_value
            )
        elif self.hypothesis == "unequal":
            return abs(statistic_value) > abs(
                self.comparison.null_dist.percentiles(100 * (1 - self.alpha / 2.0))
            )
        else:
            raise ValueError("Unknown hypothesis: {!r}".format(self.hypothesis))

    # @abstractmethod
    def _make_results(self) -> FrequentistInferenceResults:
        test_stats = self.test_stats
        accept_hypothesis = self.accept_hypothesis(test_stats["statistic_value"])

        aux = {
            "control_bootstrap_samples": self.comparison.control_bootstrap,
            "variation_bootstrap_samples": self.comparison.variation_bootstrap,
            "delta_bootstrap_samples": self.comparison.deltas_dist,
        }

        return FrequentistInferenceResults(
            control=self.comparison.d2,
            variation=self.comparison.d1,
            metric_name=self.metric_name,
            comparison_type="Bootstrap Delta",
            delta=self.comparison.delta,
            delta_relative=self.comparison.delta_relative,
            effect_size=self.comparison.effect_size,
            alpha=self.comparison.alpha,
            power=self.comparison.power,
            delta_confidence_interval=self.delta_ci,
            delta_confidence_interval_percentiles=self.delta_ci_percentiles,
            hypothesis=self.hypothesis,
            inference_method=self.inference_method,
            variable_type=self.variable_type,
            warnings=self.comparison.warnings,
            test_statistic_name=test_stats["statistic_name"],
            test_statistic_value=test_stats["statistic_value"],
            p_value=test_stats["p_value"],
            degrees_freedom=None,
            accept_hypothesis=accept_hypothesis,
            visualization_function=visualize_bootstrap_delta_results,
            aux=aux,
        )



from spearmint import table


def test_spearmint_table():
    spearmint_table = table.SpearmintTable(title="\ntest title", min_width=10)
    spearmint_table.add_column("test row", justify="right")

    # print once
    spearmint_table.print()

    # print again, with new title
    spearmint_table.title = "\nnew title"
    spearmint_table.print()
    assert len(spearmint_table._print_history) == 2

    # get printed text, excluding table grids
    first_printed_lines = spearmint_table._print_history[0].split("\n")[1::2]

    assert "test title" in first_printed_lines[0]
    assert "test row" in first_printed_lines[1]

    # get printed text, excluding table grids
    second_printed_lines = spearmint_table._print_history[1].split("\n")[1::2]
    assert "new title" in second_printed_lines[0]
    assert "test title" not in second_printed_lines[0]
    assert "test row" in second_printed_lines[1]


def test_spearmint_table_extension():
    class MySpearmintTable(table.SpearmintTable):
        def __init__(self):
            super().__init__(title="\nmy table")

            self.add_column("my first column", justify="right")
            self.add_column("my second column", justify="left")

            self.add_row("first value", "second value")

    my_table = MySpearmintTable()
    my_table.print()

    # get printed text, excluding table grids
    printed_lines = my_table._print_history[0].split("\n")[1::2]
    assert "my table" in printed_lines[0]
    assert "my first column" in printed_lines[1]
    assert "my second column" in printed_lines[1]
    assert "first value" in printed_lines[2]
    assert "second value" in printed_lines[2]



import numpy as np
import pytest

from spearmint import Experiment, HypothesisTest
from spearmint.utils import generate_fake_observations


def custom_statistic(values) -> float:
    """Silly statistic functino for testing custom boostraps"""
    return np.mean(values) ** 2


@pytest.fixture()
def proportions_data():
    return generate_fake_observations(
        distribution="bernoulli",
        n_treatments=6,
        n_attributes=1,
        n_observations=6 * 1000,
        random_seed=123,
    )


def test_small_default_bootstrap_unequal_ab_test(proportions_data):
    exp = Experiment(proportions_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        hypothesis="unequal",
        inference_method="bootstrap",
    )
    results_ab = exp.run_test(test_ab)

    assert results_ab.test_statistic_name == "bootstrap_mean"
    assert results_ab.accept_hypothesis


def test_small_default_bootstrap_unequal_aa_test(proportions_data):
    exp = Experiment(proportions_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        hypothesis="unequal",
        inference_method="bootstrap",
    )

    results_ab = exp.run_test(test_ab)

    assert results_ab.test_statistic_name == "bootstrap_mean"
    assert not results_ab.accept_hypothesis


def test_small_default_bootstrap_smaller_ab_test(proportions_data):
    exp = Experiment(proportions_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="D",
        hypothesis="smaller",
        inference_method="bootstrap",
    )
    results_ab = exp.run_test(test_ab)

    assert not results_ab.accept_hypothesis


def test_small_bootstrap_larger_ab_test(proportions_data):
    exp = Experiment(proportions_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="D",
        hypothesis="smaller",
        inference_method="bootstrap",
    )
    results_ab = exp.run_test(test_ab)

    assert not results_ab.accept_hypothesis


def test_small_custom_statistic_bootstrap_ab_test(proportions_data):
    exp = Experiment(proportions_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="D",
        hypothesis="larger",
        inference_method="bootstrap",
        statistic_function=custom_statistic,
    )
    results_ab = exp.run_test(test_ab)

    assert results_ab.test_statistic_name == "bootstrap_custom_statistic"
    assert results_ab.accept_hypothesis


def test_small_custom_statistic_bootstrap_smaller_ab_test(proportions_data):
    exp = Experiment(proportions_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="D",
        hypothesis="smaller",
        inference_method="bootstrap",
        statistic_function=custom_statistic,
    )
    results_ab = exp.run_test(test_ab)

    assert not results_ab.accept_hypothesis


@pytest.mark.filterwarnings("ignore")
def test_small_custom_statistic_bootstrap_aa_test(proportions_data):
    exp = Experiment(proportions_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        hypothesis="unequal",
        inference_method="bootstrap",
        statistic_function=custom_statistic,
    )
    # AA test results in division by zero
    results_ab = exp.run_test(test_ab)

    assert results_ab.test_statistic_name == "bootstrap_custom_statistic"
    assert not results_ab.accept_hypothesis



from rich.console import Console
from rich.table import Table as RichTable

from spearmint.typing import Optional


class SpearmintTable(RichTable):
    """Consistently-formatted Rich table with history"""

    def __init__(self, title: Optional[str] = None, **rich_table_kwargs):
        """
        Formatted table with print history

        Parameters
        ----------
        title : str, optional
            An optional title for the the able, by default None

        **rich_table_kwargs
            Any arguments that can be provided to `rich.table.Table`.` See
            https://rich.readthedocs.io/en/stable/reference/table.html# for
            details
        """
        title = title if title is not None else ""
        super().__init__(
            title=f"{title}",
            title_justify="left",
            header_style="italic cyan",
            title_style="bold cyan",
            **rich_table_kwargs,
        )
        self._print_history: list = []

    def print(self) -> None:
        """Print the table state, appending to print history"""
        console = Console(record=True)
        console.print(self)
        self._print_history.append(console.export_text())



from spearmint.experiment import Experiment
from spearmint.hypothesis_test import HypothesisTest, CustomMetric
from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("spearmint")
except PackageNotFoundError:
    # package is not installed
    pass

__all__ = ["Experiment", "HypothesisTest", "CustomMetric"]



from abc import ABC, abstractmethod

import holoviews as hv
import numpy as np
from holoviews.core.overlay import Overlay
from holoviews.element.chart import Bars, Curve
from scipy import stats

from spearmint.config import COLORS, FIGURE_PARAMS
from spearmint.typing import List, Tuple, Union

N_GRID_POINTS = 100
PDF_ZERO = 1e-5
DEFAULT_DISTRIBUTION_COLOR = COLORS.blue


class _RandomVariableClass:
    """Placeholder class for type hinting. Do not extend!"""

    def __init__(self):
        pass

    def ppf(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        return values

    def pdf(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        return values

    def pmf(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        return values

    def cdf(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        return values

    def evaluate(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        return values

    def rvs(self, size: int) -> np.ndarray:
        return np.ones(size)

    def resample(self, size: int) -> np.ndarray:
        return np.ones(size)


_RandomVariable = _RandomVariableClass()


class ProbabilityDistribution(ABC):
    """Base class for plottable probability distributions"""

    def __init__(self, label: str = "", color: str = DEFAULT_DISTRIBUTION_COLOR):
        self.label = label
        self.color = color
        self.dist = _RandomVariable

    def get_series(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get the default domain and range for plotting the probability
        distribution.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            _description_
        """
        values = self.values_grid.ravel()
        probs = self.density(values)
        return values, probs

    def sample(self, sample_size: int) -> np.ndarray:
        """
        Return a sample of `sample_size` from the probability distribution
        """
        return self.dist.rvs(size=sample_size)

    def cdf(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """
        Return the Cumulative Distribution function for the probability
        distribution, evaluated at `values`
        """
        return self.dist.cdf(values)

    def ppf(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        return self.dist.ppf(values)

    @abstractmethod
    def _get_values_grid(self) -> np.ndarray:
        """Private abstractmethod to get default values grid"""
        pass

    @property
    def values_grid(self) -> np.ndarray:
        """
        Return the default domain values for plotting
        """
        return self._get_values_grid()

    @abstractmethod
    def density(self, values: np.ndarray) -> np.ndarray:
        """
        Public abstract method to evaluate the PDF/PMF at the provided values.
        """
        pass


class Pdf(ProbabilityDistribution):
    """
    Base class for continuous probability density functions.
    """

    def __init__(self, label: str = "PDF", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label = label

    def density(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:  # type: ignore
        return self.dist.pdf(values)

    def plot(self, **plot_opts) -> Curve:  # pragma: no cover
        plot_data = self.get_series()
        plot_opts.update(FIGURE_PARAMS)
        return hv.Curve(
            data=plot_data, label=self.label, kdims="value", vdims="pdf"
        ).opts(color=self.color, **plot_opts)


class Pmf(ProbabilityDistribution):
    """
    Base class for discrete probability mass functions.
    """

    def __init__(self, label: str = "PMF", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label = label

    def density(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:  # type: ignore
        return self.dist.pmf(values)

    def plot(self, **plot_opts) -> Bars:  # pragma: no cover
        values, pmf = self.get_series()
        plot_data = zip(values, pmf)
        plot_opts.update(FIGURE_PARAMS)

        return hv.Curve(
            data=plot_data, label=self.label, kdims="value", vdims="pdf"
        ).opts(color=self.color, **plot_opts)


class ProbabilityDistributionGroup:
    """
    Convenience class for plotting a group of `ProbabilityDistribution` instances.
    """

    def __init__(self, pdfs: List[Union[Pdf, Pmf]]):
        self.pdfs = pdfs

    def plot(self, **plot_opts) -> Overlay:  # pragma: no cover
        plot_opts.update(FIGURE_PARAMS)
        overlay = self.pdfs[0].plot(**plot_opts)
        for pdf in self.pdfs[1:]:
            overlay *= pdf.plot(**plot_opts)
        return overlay


class Gaussian(Pdf):
    """
    A Gaussian probability density function
    """

    def __init__(
        self,
        mean: float = 0.0,
        std: float = 1.0,
        label: str = "Gaussian",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.label = label
        self.mean = mean
        self.std = std
        self.dist = stats.norm(loc=mean, scale=std)

    def _get_values_grid(self) -> np.ndarray:
        _min = self.ppf(PDF_ZERO)
        _max = self.ppf(1 - PDF_ZERO)
        return np.linspace(_min, _max, N_GRID_POINTS + 1)


class Bernoulli(Pmf):
    """
    A Bernoulli probability mass function
    """

    def __init__(self, p: float = 0.5, label: str = "Bernoulli", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label = label
        self.p = p
        self.dist = stats.bernoulli(p)

    def _get_values_grid(self) -> np.ndarray:
        return np.linspace(0.0, 1.0, 2)


class Binomial(Pmf):
    """
    A Binomial probability mass function
    """

    def __init__(
        self, n: int = 20, p: float = 0.5, label: str = "Binomial", *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.label = label
        self.n = n
        self.p = p
        self.dist = stats.binom(n, p)

    def _get_values_grid(self):
        _min = self.ppf(PDF_ZERO)
        _max = self.ppf(1 - PDF_ZERO)
        resolution = int(_max - _min) + 1

        if resolution > 25:
            resolution = 26

        return np.floor(np.linspace(_min, _max, resolution)).astype(int)


class Poisson(Pmf):
    """
    A Poisson probability mass function
    """

    def __init__(self, mu: float = 1.0, label: str = "Poisson", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label = label
        self.mu = mu
        self.dist = stats.poisson(mu)

    def _get_values_grid(self):
        _min = self.ppf(PDF_ZERO)
        _max = self.ppf(1 - PDF_ZERO)
        return np.arange(_min, _max + 1)


class Kde(Pdf):
    """
    Estimate the shape of a PDF using a Gaussian kernel density estimator.
    """

    def __init__(self, samples: np.ndarray, label: str = "KDE", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label = label
        self.dist = stats.gaussian_kde(samples)
        self.low = min(samples)
        self.high = max(samples)

    def _get_values_grid(self):
        return np.linspace(self.low, self.high, N_GRID_POINTS + 1)

    def density(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:  # type: ignore
        """Note, we overload Pdf.density here"""
        return self.dist.evaluate(values)

    def sample(self, sample_size: int) -> np.ndarray:
        """Note: we sample 1D arrays"""
        return self.dist.resample(size=sample_size).flatten()



import json

import pytest

from spearmint import mixin


def test_jsonable():
    class TestJsonable(mixin.JsonableMixin):
        def __init__(self):
            self.abra = "cadabra"

    ts = TestJsonable()
    ts_dict = ts.to_dict()
    assert "abra" in ts_dict
    assert ts_dict["abra"] == "cadabra"
    assert json.loads(ts.__repr__()) == ts_dict


def test_dataframeable():
    class ValidDataFrameable(mixin.DataframeableMixin):
        def to_dict(self):
            return {"abra": ("cadabra", "calamazam"), "spearmint": "refreshing"}

    class InvalidDataframeable(mixin.DataframeableMixin):
        pass

    with pytest.raises(NotImplementedError):
        InvalidDataframeable().to_dataframe()

    vdf = ValidDataFrameable()
    vdf_dict = vdf.to_dict()

    assert vdf_dict == {"abra": ("cadabra", "calamazam"), "spearmint": "refreshing"}
    df = vdf.to_dataframe()
    assert df.columns.tolist() == ["abra", "spearmint"]
    assert df.iloc[0].tolist() == [("cadabra", "calamazam"), "refreshing"]



import numpy as np
from scipy import stats

from spearmint import distributions as dst

TEST_LABEL = "my_pdf"
TEST_COLOR = "blue"
TEST_VALUES = np.array([1, 1])


class MockDist:
    def rvs(self, size):
        return np.ones(size)

    def pdf(self, values):
        return np.ones(len(values))

    def pmf(self, values):
        return np.ones(len(values))

    def ppf(self, values):
        return np.ones(len(values))

    def cdf(self, values):
        return np.ones(len(values))


class ExtendedProbabilityDistribution(dst.ProbabilityDistribution):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # This distribution will alwiays return 0. rvs
        self.dist = MockDist()

    def _get_values_grid(self):
        return TEST_VALUES

    def density(self, values):
        return TEST_VALUES


class ExtendedPdf(ExtendedProbabilityDistribution):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ExtendedPmf(ExtendedProbabilityDistribution):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def test_probability_distribution():
    pdf = ExtendedProbabilityDistribution(label=TEST_LABEL, color=TEST_COLOR)
    assert pdf.label == TEST_LABEL
    assert pdf.color == TEST_COLOR
    assert np.all(pdf.values_grid == TEST_VALUES)
    assert np.all(pdf.density(TEST_VALUES) == TEST_VALUES)

    vals, probs = pdf.get_series()
    assert np.all(vals == TEST_VALUES)
    assert np.all(probs == TEST_VALUES)

    assert np.all(pdf.sample(3) == np.array([1, 1, 1]))
    assert np.all(pdf.ppf(np.array([0])) == np.array([1]))


def test_pdf():
    pdf = ExtendedPdf(label=TEST_LABEL, color=TEST_COLOR)
    assert pdf.label == TEST_LABEL
    assert pdf.color == TEST_COLOR

    assert np.all(pdf.values_grid == TEST_VALUES)
    assert np.all(pdf.density(TEST_VALUES) == TEST_VALUES)

    vals, probs = pdf.get_series()
    assert np.all(vals == TEST_VALUES)
    assert np.all(probs == TEST_VALUES)

    assert np.all(pdf.sample(3) == np.array([1, 1, 1]))
    assert np.all(pdf.ppf(np.array([0])) == np.array([1]))

    # Pdf uses dist.pdf
    assert np.all(pdf.density(TEST_VALUES) == MockDist().pdf(TEST_VALUES))


def test_pmf():
    pmf = ExtendedPmf(label=TEST_LABEL, color=TEST_COLOR)
    assert pmf.label == TEST_LABEL
    assert pmf.color == TEST_COLOR

    assert np.all(pmf.values_grid == TEST_VALUES)
    assert np.all(pmf.density(TEST_VALUES) == TEST_VALUES)

    vals, probs = pmf.get_series()
    assert np.all(vals == TEST_VALUES)
    assert np.all(probs == TEST_VALUES)

    assert np.all(pmf.sample(3) == np.array([1, 1, 1]))
    assert np.all(pmf.ppf(np.array([0])) == np.array([1]))

    # Pmf uses dist.pmf
    assert np.all(pmf.density(TEST_VALUES) == MockDist().pmf(TEST_VALUES))


def test_probability_distribution_group():
    pmf = ExtendedPmf(label=TEST_LABEL, color=TEST_COLOR)
    pdf = ExtendedPdf(label=TEST_LABEL, color=TEST_COLOR)

    pdg = dst.ProbabilityDistributionGroup(pdfs=[pmf, pdf])

    for pd in pdg.pdfs:
        assert isinstance(pd, dst.ProbabilityDistribution)


def test_kde():
    np.random.seed(123)
    samples = np.random.randn(10000)
    kde = dst.Kde(samples=samples, label=TEST_LABEL, color=TEST_COLOR)
    values_grid = kde.values_grid
    assert values_grid.min() == samples.min()
    assert values_grid.max() == samples.max()
    assert kde.density(-100) == 0
    assert kde.density(100) == 0

    kde_samples = kde.sample(3)
    assert isinstance(kde_samples, np.ndarray)
    assert len(kde_samples) == 3


def test_gaussian_pdf():
    MEAN = 1
    STD = 2
    pdf = dst.Gaussian(mean=MEAN, std=STD, label=TEST_LABEL, color=TEST_COLOR)

    assert pdf.mean == MEAN
    assert pdf.std == STD
    assert pdf.label == TEST_LABEL
    assert pdf.color == TEST_COLOR

    grid_vals = pdf._get_values_grid()
    norm = stats.norm(MEAN, STD)
    assert grid_vals[0] == norm.ppf(dst.PDF_ZERO)
    assert grid_vals[-1] == norm.ppf(1 - dst.PDF_ZERO)


def test_bernoulli_pmf():
    P = 0.5
    pdf = dst.Bernoulli(p=P, label=TEST_LABEL, color=TEST_COLOR)

    assert pdf.p == P
    assert pdf.label == TEST_LABEL
    assert pdf.color == TEST_COLOR

    grid_vals = pdf._get_values_grid()
    assert grid_vals[0] == 0
    assert grid_vals[-1] == 1


def test_binomial_pmf():
    P = 0.5
    N = 10
    pdf = dst.Binomial(p=P, n=N, label=TEST_LABEL, color=TEST_COLOR)

    assert pdf.p == P
    assert pdf.n == N
    assert pdf.label == TEST_LABEL
    assert pdf.color == TEST_COLOR

    grid_vals = pdf._get_values_grid()
    binom = stats.binom(N, P)
    assert grid_vals[0] == binom.ppf(dst.PDF_ZERO)
    assert grid_vals[-1] == binom.ppf(1 - dst.PDF_ZERO)


def test_poisson_pmf():
    MU = 1
    pdf = dst.Poisson(mu=MU, label=TEST_LABEL, color=TEST_COLOR)

    assert pdf.mu == MU
    assert pdf.label == TEST_LABEL
    assert pdf.color == TEST_COLOR

    grid_vals = pdf._get_values_grid()
    poisson = stats.poisson(MU)
    assert grid_vals[0] == poisson.ppf(dst.PDF_ZERO)
    assert grid_vals[-1] == poisson.ppf(1 - dst.PDF_ZERO)



from spearmint import vis


def test_colors():
    assert hasattr(vis.COLORS, "blue")
    assert hasattr(vis, "DEFAULT_COLOR")
    assert hasattr(vis, "VARIATION_COLOR")
    assert hasattr(vis, "CONTROL_COLOR")
    assert hasattr(vis, "DELTA_COLOR")
    assert hasattr(vis, "POSITIVE_COLOR")
    assert hasattr(vis, "NEGATIVE_COLOR")
    assert hasattr(vis, "NEUTRAL_COLOR")



import pytest


def pytest_addoption(parser):
    """
    https://docs.pytest.org/en/7.1.x/example/markers.html
    """
    parser.addoption(
        "--include_pymc_tests",
        action="store_true",
        default=False,
        help="Run Bayesian tests that require PyMC, which are skipped by default. ",
    )


def pytest_collection_modifyitems(config, items):
    """
    https://docs.pytest.org/en/7.1.x/example/markers.html
    """
    if not config.getoption("--include_pymc_tests"):
        skip_slow_tests = pytest.mark.skip(
            reason="Skipping Bayesian tests that require PyMC - use --include_pymc_tests to run"
        )
        for item in items:
            if "pymc_test" in item.keywords:
                item.add_marker(skip_slow_tests)



import pytest

from spearmint import config
from spearmint import Experiment, HypothesisTest
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def continuous_data():
    return generate_fake_observations(
        distribution="gaussian", n_treatments=2, n_observations=2 * 100, random_seed=123
    )


@pytest.mark.pymc_test
def test_bayesian_continuous_default(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric", control="A", variation="B", inference_method="bayesian"
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert test_results.model_name == "gaussian"
    assert (
        test_results.bayesian_parameter_estimation_method
        == config.DEFAULT_PARAMETER_ESTIMATION_METHOD
    )
    assert test_results.model_hyperparams["prior_mean_mu"] == 0.0
    assert test_results.model_hyperparams["prior_var_mu"] == 5.0

    if config.DEFAULT_PARAMETER_ESTIMATION_METHOD in ("mcmc", "advi"):
        assert test_results.model_hyperparams["prior_mean_sigma"] == 1.0
        assert test_results.model_hyperparams["prior_var_sigma"] == 5.0

    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bayesian_gaussian_ab_mcmc(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        inference_method="bayesian",
        # bayesian_parameter_estimation_method="mcmc",  # MCMC is default
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert (
        test_results.model_name == "gaussian"
    )  # Gaussian is default for continuous data
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bayesian_gaussian_aa_mcmc(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        # bayesian_parameter_estimation_method="mcmc",  # MCMC is default
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert (
        test_results.model_name == "gaussian"
    )  # Gaussian is default for continuous data
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


@pytest.mark.pymc_test
def test_bayesian_gaussian_ab_advi(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        inference_method="bayesian",
        bayesian_parameter_estimation_method="advi",  # use ADVI parameter estimation
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert (
        test_results.model_name == "gaussian"
    )  # Gaussian is default for continuous data
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bayesian_gaussian_aa_advi(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_parameter_estimation_method="advi",  # use ADVI parameter estimation
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert (
        test_results.model_name == "gaussian"
    )  # Gaussian is default for continuous data
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


def test_bayesian_gaussian_ab_analytic(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        inference_method="bayesian",
        bayesian_parameter_estimation_method="analytic",  # use Analytic parameter estimation
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert (
        test_results.model_name == "gaussian"
    )  # Gaussian is default for continuous data
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


def test_bayesian_gaussian_aa_analytic(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_parameter_estimation_method="analytic",  # use Analytic parameter estimation
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert (
        test_results.model_name == "gaussian"
    )  # Gaussian is default for continuous data
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


@pytest.mark.pymc_test
def test_bayesian_student_t_ab_mcmc(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        inference_method="bayesian",
        bayesian_model_name="student_t",
        bayesian_parameter_estimation_method="mcmc",  # MCMC is default
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert test_results.model_name == "student_t"
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bayesian_student_t_aa_mcmc(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_model_name="student_t",
        bayesian_parameter_estimation_method="mcmc",
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert test_results.model_name == "student_t"
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


@pytest.mark.pymc_test
def test_bayesian_student_t_ab_advi(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        inference_method="bayesian",
        bayesian_model_name="student_t",
        bayesian_parameter_estimation_method="advi",  # use ADVI parameter estimation
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert test_results.model_name == "student_t"
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bayesian_student_t_aa_advi(continuous_data):
    exp = Experiment(data=continuous_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_model_name="student_t",
        bayesian_parameter_estimation_method="advi",  # use ADVI parameter estimation
    )

    test_results = exp.run_test(test)
    test_results.display()

    assert test_results.model_name == "student_t"
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )



import pymc as pm
from scipy import stats

from spearmint.stats import Samples
from spearmint.typing import Dict, Tuple

from .analytic_base import BayesianAnalyticModel


def _get_gamma_prior_params(samples: Samples) -> Tuple[float, float]:
    """Estimate the prior parameters for Beta distribution from samples"""
    alpha = samples.mean**2 / samples.var
    beta = samples.mean / samples.var

    return alpha, beta


class PoissonAnalyticModel(BayesianAnalyticModel):
    """
    Implement analytic posterior updates for Gamma-Poisson model. Namely,
    a Gamma prior and Poisson likelihood result in a Gamma posterior over
    rate parameters, which can be calculated efficiently from the prior and
    descriptive statistics of the observations.

    References
    ----------
    - TBD
    """

    def __init__(
        self, prior_alpha: float = 1, prior_beta: float = 1.0, *args, **kwargs
    ):
        """
        Parameters
        ----------
        prior_alpha : float, optional
            The shape parameter for the Gamma prior distribution
        prior_beta : float, optional
            The shape parametrer for the Gamma prior distribution
        """
        super().__init__(delta_param="lambda", *args, **kwargs)  # type: ignore # (mypy bug, see #6799)
        self.prior_alpha = prior_alpha
        self.prior_beta = prior_beta

    def calculate_posteriors(
        self, control_samples: Samples, variation_samples: Samples
    ) -> None:
        """
        Update the posterior distributions for the control and variation in
        light of Samples

        Parameters
        ----------
        control_samples : Samples
            Observations for the control group
        variation_samples : Samples
            Observations for the variation group
        """

        # Update rules
        def _posterior_alpha(prior_alpha, samples):
            return prior_alpha + samples.sum

        def _posterior_beta(prior_beta, samples):
            return prior_beta + samples.nobs

        control_posterior_alpha = _posterior_alpha(self.prior_alpha, control_samples)
        control_posterior_beta = _posterior_beta(self.prior_beta, control_samples)

        variation_posterior_alpha = _posterior_alpha(
            self.prior_alpha, variation_samples
        )
        variation_posterior_beta = _posterior_beta(self.prior_beta, variation_samples)

        # Add prior for visualization
        self._prior = stats.gamma(a=self.prior_alpha, scale=1 / self.prior_beta)
        self._control_posterior = stats.gamma(
            a=control_posterior_alpha, scale=1 / control_posterior_beta
        )

        self._variation_posterior = stats.gamma(
            a=variation_posterior_alpha, scale=1 / variation_posterior_beta
        )


def build_poisson_analytic_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_alpha: float = 1.0,
    prior_beta: float = 1.0,
) -> Tuple[PoissonAnalyticModel, Dict[str, float]]:
    model = PoissonAnalyticModel(prior_alpha=prior_alpha, prior_beta=prior_beta)
    model.calculate_posteriors(control_samples, variation_samples)

    hyperparams = {"prior_alpha": prior_alpha, "prior_beta": prior_beta}
    return model, hyperparams


def build_poisson_pymc_model(
    control_samples: Samples,
    variation_samples: Samples,
    prior_alpha: float = 1.0,
    prior_beta: float = 1.0,
) -> pm.Model:
    """
    Compile a Gamma-Poisson Bayesian PyMC model for modeling counted events data.
    The model consists of a Gamma prior over event rate, and a Poisson likelihood.
    For the Gamma prior, we derive parameters separately for the control and
    variation from their observations (i.e. no pooling).

    Parameters
    ----------
    control_observations: Samples, dtype=int
        The control group observations
    variation_observations: Samples, dtype=int
        The variation group observations

    Returns
    -------
    model : pm.Model
        The compiled PyMC model built with the provided data
    hyperparams : Dict
        The prior parameters derived from the Samples

    References
    ----------
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Gamma.html
    -   https://www.pymc.io/projects/docs/en/stable/api/distributions/generated/pymc.Poisson.html
    """

    hyperparams = {"prior_alpha": prior_alpha, "prior_beta": prior_beta}

    with pm.Model() as model:
        # Priors
        pm.Gamma("prior", alpha=prior_alpha, beta=prior_beta)  # for visualizing prior
        lambda_control = pm.Gamma("lambda_control", alpha=prior_alpha, beta=prior_beta)
        lambda_variation = pm.Gamma(
            "lambda_variation", alpha=prior_alpha, beta=prior_beta
        )

        # Likelihoods
        pm.Poisson("control", mu=lambda_control, observed=control_samples.data)
        pm.Poisson("variation", mu=lambda_variation, observed=variation_samples.data)

        # Inference parameters
        pm.Deterministic("delta", lambda_variation - lambda_control)
        effect_size = pm.Deterministic(
            "effect_size", (lambda_variation / lambda_control) - 1.0
        )
        pm.Deterministic("delta_relative", effect_size - 1)

    return model, hyperparams



import os
import shutil
import numpy as np
import pandas as pd
import string

from spearmint.typing import Any, List, Iterable, Union, DataFrame, Path, Optional

BIG_FLOAT = 2.0**32
SMALL_FLOAT = -(2.0**32)
NAN_TYPE_MAPPING = {
    np.inf: None,
    -np.inf: None,
    np.nan: None,
}


def mkdir(dirname: Union[str, Path]) -> None:
    """Make directory, including full file path

    Parameters
    ----------
    dirname : Union[str, Path]
        The fullfile path to make

    Raises
    ------
    OSError if any trouble creating the directory
    """
    if not os.path.isdir(dirname):
        os.makedirs(dirname)


def rmdir(dirname: Union[str, Path]) -> None:
    """Remove the fullpath to the directory

    Parameters
    ----------
    dirname : Union[str, Path]
        The fullfile path to remove

    Raises
    ------
    OSError if any trouble creating the directory
    """
    if os.path.isdir(dirname):
        shutil.rmtree(dirname)


def process_warnings(warnings: Union[str, List[str], None]) -> str:
    """Convert sequence of warnings into a string"""
    if warnings is None:
        return ""

    def _flatten(_warnings):  # noqa
        for w in _warnings:
            if isinstance(w, Iterable) and not isinstance(w, str):
                for subw in _flatten(w):
                    yield subw
            else:
                yield w

    if len(warnings) > 0:
        if isinstance(warnings, list):
            flattened_warnings = _flatten(warnings)
            warnings = "; ".join(flattened_warnings)
        return warnings
    return ""


def isnumeric(val: Any) -> bool:
    try:
        float(val)
        return True
    except ValueError:
        return False


def coerce_value(val: Any) -> Any:
    """
    Infer and cast value to valid types

    Parameters
    ----------
    val : Any
        The value to coerce

    Returns
    -------
    coerced_value : Any
        The value with inferred type

    Raises
    ------
    ValueError if can't infer the type from the value.
    """

    if val in NAN_TYPE_MAPPING:
        return NAN_TYPE_MAPPING[val]
    if isnumeric(val):
        try:
            return int(val)
        except ValueError:
            return float(val)

    lower_val = str(val.lower())
    if lower_val in ("true", "false"):
        if "f" in lower_val:
            return False
        else:
            return True

    if "," in val:
        return [coerce_value(v.strip()) for v in val.split(",")]
    return val


def format_value(
    val: Optional[Union[float, Iterable[float]]], precision: int = 4
) -> str:
    """Helper function for standardizing the precision of numerical and
    iterable values, returning a string representation.

    Parameters
    ----------
    val : Union[float, Iterable[float]]
        The value to format into a string
    precision : int
        The output numerical precision

    Returns
    -------
    formatted_value : str
        Representation of `val` with proper precision
    """
    if val is None:
        return ""

    if isinstance(val, Iterable):
        return f"{tuple((round(v, precision) for v in val))}"
    if precision == 0:
        return f"{int(val)}"
    return f"{round(val, precision)}"


def ensure_dataframe(data: Any, data_attr: str = "data") -> DataFrame:
    """
    Check if an object is a dataframe, and if not, check if it has an
    attribute that is a dataframe, and return that attribute

    Parameters
    ----------
    data : Union[DataFrame, object]
        The data to check

    Returns
    -------
    data : DataFrame
        A verified dataframe

    Raises
    ------
    - ValueError if data `data.data_attr` isn't a DataFrame
    """
    if not isinstance(data, DataFrame):
        if hasattr(data, data_attr) and isinstance(getattr(data, data_attr), DataFrame):
            data = getattr(data, data_attr)
        else:
            raise ValueError("`data` is incorrect format, must be a DataFrame")

    return data


def infer_variable_type(values: np.ndarray) -> str:
    """Infer the type of variable distribution from `values`

    Parameters
    ----------
    values : np.ndarray
        An array of observations

    Returns
    -------
    str
        The type of variable. Either 'binary', 'counts', or 'continuous'.
    """
    if values.dtype == bool:
        return "binary"
    if len(set(values)) == 2 and values.min() == 0 and values.max() == 1:
        return "binary"
    if sum(values.astype(int) - values) == 0:
        return "counts"
    return "continuous"


def set_matplotlib_backend() -> str:
    """
    Set supported matplotlb backend depending on the current platform.

    Returns
    -------
    backend : str
        The matplotlib backend being used
    """
    from sys import platform
    import matplotlib as mpl

    backend = "pdf" if platform == "darwin" else "agg"
    mpl.use(backend)
    return backend


def safe_isnan(val: Any) -> bool:
    """
    Check if a value is a NaN, handling `None` values

    Parameters
    ----------
    val : Any
        A value to check

    Returns
    -------
    is_nan : bool
        Truth value of the the value being a NaN
    """
    if val is not None:
        return np.isnan(val)  # type: ignore
    return False


def generate_fake_observations(
    n_observations: int = 10000,
    n_treatments: int = 2,
    n_attributes: int = 2,
    distribution: str = "bernoulli",
    random_seed: int = 123,
) -> DataFrame:
    """
    Create a dataframe of artificial observations to be used for testing and demos.
    Treatments have different means, but segments defined by the attributes have
    no effect.

    Parameters
    -----------
    n_observations: int
        number of unique observations (e.g. users)
    n_treatments: int
        the number of simulated treatments. Will create an equivalent
        number of cohorts. Each successive treatment, by definition will
        have a systematically larger mean metric values (see metric
        distribution below). Note: the maximum number of treatments allowed
        is 6.
    n_attributes: int
        the number of attribute columns simulated. The number of distinct
        values taken by each attribute column is sampled uniformly between
        1 and 4. The segments defined by n_attributes have no effect on the
        sample means.
    distribution: str
        the type of metric distributions simulated
            -   'bernoulli': binary metric values with mean proporionality increasing
                from .5 by .1 for each successive treatment
            -   'gaussian': continuous metric values with the mean increasing
                from 0 by 1 for each successive treatment
            -   'poisson': counts metric values with the mean increasing from 1
                by 1 for each successive treatment
    seed: int
        The random number generator seed.

    Returns
    -------
    fake_observations : DataFrame
        Synthethic dataset with the columns:
        -   'metric': metric with support of `distribution`
        -   'treatment': the alphabetic label of treatments (e.g. 'A', 'B', etc.)
        -   'attr_*': randomly-genrated on-hot encoded attributes associated with
            each synthetic observation
    """
    distribution_ = distribution.lower()
    if distribution_ not in ("poisson", "bernoulli", "gaussian"):
        raise ValueError(f"Unsupported distribution {distribution}")
    np.random.seed(random_seed)

    letters = string.ascii_uppercase
    n_treatments = min(n_treatments, 6)

    data = pd.DataFrame()
    data["id"] = list(range(n_observations))

    # Add treatments
    treatments = list(letters[:n_treatments])
    data["treatment"] = np.random.choice(treatments, size=n_observations)

    # Add attributes (attributes should have no effect)
    attribute_columns = ["attr_{}".format(i) for i in range(n_attributes)]
    for ai, attr in enumerate(attribute_columns):
        attr_vals = [
            "A{}{}".format(ai, a.lower())
            for a in list(letters[: np.random.randint(1, 4)])
        ]
        data[attr] = np.random.choice(attr_vals, size=n_observations)

    # Add measurements, each treatment has successively larger means
    for delta, tr in enumerate(treatments):
        tr_mask = data.treatment == tr
        n_tr = sum(tr_mask)
        if distribution_ == "gaussian":
            data.loc[tr_mask, "metric"] = delta + np.random.randn(n_tr)
        elif distribution_ == "bernoulli":
            data.loc[tr_mask, "metric"] = np.round(0.1 * delta + np.random.random(n_tr))
        elif distribution_ == "poisson":
            data.loc[tr_mask, "metric"] = np.random.poisson(1 + delta, size=n_tr)

    # Set the metric data type
    if distribution_ == "poisson":
        dtype = int  # type: ignore
    elif distribution_ == "bernoulli":
        dtype = bool  # type: ignore
    elif distribution_ == "gaussian":
        dtype = float  # type: ignore

    data = data.astype({"metric": dtype})

    # reorder columns
    base_cols = ["id", "treatment", "metric"]
    attr_cols = [c for c in data.columns if c not in base_cols]
    reordered_cols = base_cols + attr_cols

    return data[reordered_cols]



import pytest

from spearmint.dataset import DataFrame, Dataset, DatasetException, search_config
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def test_observations():
    return generate_fake_observations(distribution="bernoulli")


def test_search_config():
    df = generate_fake_observations(n_observations=1)

    # Test against default config template
    assert (
        "treatment" in search_config(df, "hypothesis_test", "default_treatment_name")[0]
    )
    assert "metric" in search_config(df, "hypothesis_test", "default_metric_name")
    assert "attr_0" in search_config(df, "hypothesis_test", "default_attribute_names")
    assert "attr_1" in search_config(df, "hypothesis_test", "default_attribute_names")


def test_default_init(test_observations):
    dataset = Dataset(test_observations)
    # default configuration template
    assert dataset.treatment == "treatment"
    assert "metric" in dataset.measures
    assert (
        dataset.__repr__()
        == "Dataset(treatment='treatment', measures=['metric'], attributes=['attr_0', 'attr_1'])"
    )


def test_properties(test_observations):
    dataset = Dataset(test_observations)

    selected_treatments = ["A", "B"]
    mask = test_observations.treatment.isin(selected_treatments)
    dataset = Dataset(test_observations[mask])
    assert dataset.cohorts == selected_treatments

    cohort_measures = dataset.cohort_measures
    assert isinstance(cohort_measures, DataFrame)
    assert cohort_measures.shape[0] == len(dataset.cohorts)


def test_segments(test_observations):
    dataset = Dataset(df=test_observations, attributes=["attr_0", "attr_1"])

    segments = dataset.segments("attr_0")
    assert isinstance(segments, list)
    assert ("A", "A0a") in segments
    assert ("B", "A0b") in segments

    segment_samples = dataset.segment_samples("attr_0")
    assert isinstance(segment_samples, DataFrame)
    assert ("A", "A0a") in segment_samples.index
    assert ("B", "A0b") in segment_samples.index

    # Replace column names with spaced column names, ensure that
    # segmentation based off pandas.query still works as exptected
    test_observations.columns = [c.replace("_", " ") for c in test_observations.columns]
    dataset = Dataset(df=test_observations, attributes=["attr 0", "attr 1"])

    segments = dataset.segments("attr 0")
    assert isinstance(segments, list)
    assert ("A", "A0a") in segments
    assert ("B", "A0b") in segments

    segment_samples = dataset.segment_samples("attr 0")
    assert isinstance(segment_samples, DataFrame)
    assert ("A", "A0a") in segment_samples.index
    assert ("B", "A0b") in segment_samples.index


def test_exceptions(test_observations):
    with pytest.raises(DatasetException):
        Dataset(test_observations, treatment="unknown treatment")

    with pytest.raises(DatasetException):
        Dataset(test_observations, measures="unknown measures string")

    with pytest.raises(DatasetException):
        Dataset(test_observations, measures=["unknown measures", "list"])

    with pytest.raises(DatasetException):
        Dataset(test_observations, attributes="unknown attribures string")

    with pytest.raises(DatasetException):
        Dataset(test_observations, attributes=["unknown attributes", "list"])

    with pytest.raises(DatasetException):
        Dataset(test_observations, metadata="unknown metadata string")

    with pytest.raises(DatasetException):
        Dataset(test_observations, metadata=["unknown metadata", "list"])



import pytest

from spearmint import config
from spearmint import Experiment, HypothesisTest
from spearmint.inference.bayesian.bayesian_inference import (
    UnsupportedParameterEstimationMethodException,
)
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def binary_data():
    return generate_fake_observations(
        distribution="bernoulli",
        n_treatments=3,
        n_observations=3 * 100,
        random_seed=123,
    )


@pytest.mark.pymc_test
def test_bayesian_binary_default(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
    )
    test_results = exp.run_test(test)

    assert test_results.model_name == "binomial"  # Default model for binary data
    assert (
        test_results.bayesian_parameter_estimation_method
        == config.DEFAULT_PARAMETER_ESTIMATION_METHOD
    )  # Default param estimation
    assert test_results.model_hyperparams["prior_alpha"] == 1.0  # Default priors
    assert test_results.model_hyperparams["prior_beta"] == 1.0


@pytest.mark.pymc_test
def test_bernoulli_mcmc_model_params(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayesian_model_name="bernoulli",
        bayesian_model_params=dict(prior_alpha=2.0),
        bayesian_parameter_estimation_method="mcmc",
    )
    test_results = exp.run_test(test)

    assert test_results.model_name == "bernoulli"
    assert test_results.model_hyperparams["prior_alpha"] == 2
    assert test_results.model_hyperparams["prior_beta"] == 1  # default


@pytest.mark.pymc_test
def test_bernoulli_ab_mcmc(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayesian_model_name="bernoulli",
        # bayesian_parameter_estimation_method="mcmc"  # Default
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "bernoulli"
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bernoulli_aa_mcmc(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_model_name="bernoulli",
        # bayesian_parameter_estimation_method="mcmc"  # Default
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "bernoulli"
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


@pytest.mark.pymc_test
def test_bernoulli_ab_advi(binary_data):
    exp = Experiment(data=binary_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayesian_model_name="bernoulli",
        bayesian_parameter_estimation_method="advi",
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "bernoulli"
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_bernoulli_aa_advi(binary_data):
    exp = Experiment(data=binary_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_model_name="bernoulli",
        bayesian_parameter_estimation_method="advi",
    )
    test_results = exp.run_test(test)

    test_results.display()
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


def test_bernoulli_analytic_model_params(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayesian_model_name="bernoulli",
        bayesian_model_params=dict(prior_alpha=2.0),
        bayesian_parameter_estimation_method="analytic",
    )
    test_results = exp.run_test(test)

    assert test_results.model_name == "bernoulli"
    assert test_results.model_hyperparams["prior_alpha"] == 2
    assert test_results.model_hyperparams["prior_beta"] == 1  # default


def test_bernoulli_aa_analytic(binary_data):
    exp = Experiment(data=binary_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_model_name="bernoulli",
        bayesian_parameter_estimation_method="analytic",
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "bernoulli"
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


def test_bernoulli_ab_analytic(binary_data):
    exp = Experiment(data=binary_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        inference_method="bayesian",
        bayesian_model_name="bernoulli",
        bayesian_parameter_estimation_method="analytic",
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "bernoulli"
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_binomial_mcmc_model_params(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayesian_parameter_estimation_method="mcmc",
        bayesian_model_params=dict(prior_alpha=2.0, possible_outcomes=3.0),
    )
    test_results = exp.run_test(test)

    assert test_results.model_name == "binomial"  # Binomial is default model for binary
    assert test_results.model_hyperparams["prior_alpha"] == 2
    assert test_results.model_hyperparams["possible_outcomes"] == 3
    assert test_results.model_hyperparams["prior_beta"] == 1  # default


@pytest.mark.pymc_test
def test_binomial_ab_mcmc(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayesian_model_name="binomial"
        # bayesian_parameter_estimation_method="mcmc"  # Default
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "binomial"  # Binomial is default model for binary
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0


@pytest.mark.pymc_test
def test_binomial_aa_mcmc(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_model_name="binomial",
        # bayesian_parameter_estimation_method="mcmc"  # Default
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "binomial"  # Binomial is default model for binary
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


@pytest.mark.pymc_test
def test_binomial_advi(binary_data):
    """
    ADVI parameter estimation not supported for discrete PDFs like the
    binomial.
    """
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayesian_model_name="binomial",
        bayesian_parameter_estimation_method="advi",
    )

    with pytest.raises(UnsupportedParameterEstimationMethodException):
        exp.run_test(test)


def test_binomial_analytic_model_params(binary_data):
    exp = Experiment(data=binary_data)

    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="C",
        inference_method="bayesian",
        bayeisan_model_name="binomial",
        bayesian_parameter_estimation_method="analytic",
        bayesian_model_params=dict(prior_alpha=2.0),
    )
    test_results = exp.run_test(test)

    assert test_results.model_name == "binomial"  # Binomial is default model for binary
    assert test_results.model_hyperparams["prior_alpha"] == 2
    assert test_results.model_hyperparams["prior_beta"] == 1  # default


def test_binomial_aa_analytic(binary_data):
    exp = Experiment(data=binary_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="bayesian",
        bayesian_model_name="binomial",
        bayesian_parameter_estimation_method="analytic",
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "binomial"  # Binomial is default model for binary
    assert not test_results.accept_hypothesis
    assert (
        not pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
    )


def test_binommial_ab_analytic(binary_data):
    exp = Experiment(data=binary_data)
    test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        inference_method="bayesian",
        bayesian_model_name="binomial",
        bayesian_parameter_estimation_method="analytic",
    )
    test_results = exp.run_test(test)

    test_results.display()

    assert test_results.model_name == "binomial"  # Binomial is default model for binary
    assert test_results.accept_hypothesis
    assert pytest.approx(test_results.prob_greater_than_zero, rel=0.1, abs=0.01) == 1.0
